#!/usr/bin/env python3
"""
Syncthing-GTK - DaemonSettingsDialog

Universal dialog handler for all Syncthing settings and editing
"""

from gi.repository import Gtk
from syncthing_gtk.tools import (
	IS_UNITY, IS_GNOME, IS_WINDOWS,
	set_run_on_startup, is_ran_on_startup, get_config_dir, get_executable
)
from syncthing_gtk.notifications import Notifications, HAS_DESKTOP_NOTIFY
from syncthing_gtk.editordialog import EditorDialog
from syncthing_gtk.tools import _ # gettext function
from syncthing_gtk.configuration import LONG_AGO
try:
	from syncthing_gtk.stdownloader import StDownloader
except ImportError:
	StDownloader = None
import os, logging

log = logging.getLogger("UISettingsDialog")

VALUES = [ "vautostart_daemon", "vautokill_daemon", "vminimize_on_start",
		"vautostart", "vuse_old_header",
		"vforce_dark_theme", "vdaemon_priority", "vfolder_as_path",
		"vnotification_for_update", "vnotification_for_folder",
		"vnotification_for_error", "vst_autoupdate", "vsyncthing_binary",
		"vsyncthing_arguments", "vmax_cpus", "vicon_theme", "vlanguage"
	]

# Values for filemanager integration. Key is ID of checkbox widget
FM_DATA = {
	"fmcb_nemo" : (
		"nemo/extensions-3.0/libnemo-python.so",	# python plugin location, relative to /usr/lib
		"Nemo python bindings",						# name or description of required package
		"syncthing-plugin-nemo",					# plugin script filename, without extension
		"nemo-python/extensions",					# script folder, relative to XDG_DATA_HOME
		"Nemo"										# name
	),
	"fmcb_nautilus" : (
		"nautilus/extensions-3.0/libnautilus-python.so",
		"Nautilus python bindings",
		"syncthing-plugin-nautilus",
		"nautilus-python/extensions",
		"Nautilus"
	),
	"fmcb_caja" : (
		"caja/extensions-2.0/libcaja-python.so",
		"Caja python bindings",
		"syncthing-plugin-caja",
		"caja-python/extensions",
		"Caja"
	)
}

class UISettingsDialog(EditorDialog):
	SETTING_NEEDS_RESTART = [
		"vuse_old_header", "vforce_dark_theme",
		"vicon_theme", "vlanguage"
	]
	
	def __init__(self, app):
		EditorDialog.__init__(self, app, "ui-settings.glade",
			_("UI Settings"))
		self.app = app
	
	def run(self):
		return self["dialog"].run()
	
	def cb_btBrowse_clicked(self, *a):
		""" Display file browser dialog to browse for syncthing binary """
		browse_for_binary(self["editor"], self, "vsyncthing_binary")
	
	def cb_vmax_cpus_value_changed(self, sb):
		if sb.get_adjustment().get_value() == 0:
			sb.set_text(_("Unlimited"))
	
	#@Overrides
	def load_data(self):
		# Don't load data from syncthing daemon, it knows nothing...
		copy = { k : self.app.config[k] for k in self.app.config }
		if IS_UNITY or IS_GNOME:
			self["vuse_old_header"].set_visible(False)
			self["vuse_old_header"].set_no_show_all(True)
		if IS_GNOME:
			self["vautostart"].set_visible(False)
			self["vautostart"].set_no_show_all(True)
		if not HAS_DESKTOP_NOTIFY:
			# Disable notifications settings if required
			# library is not available
			self["lblNotifications"].set_sensitive(False)
			self["vnotification_for_update"].set_sensitive(False)
			self["vnotification_for_folder"].set_sensitive(False)
			self["vnotification_for_error"].set_sensitive(False)
		if IS_WINDOWS:
			# Leave daemon running causes weird bugs on Windows,
			# so only one option is enabled there
			self["rbOnExitLeave"].set_sensitive(False)
			self["rbOnExitAsk"].set_sensitive(False)
			self["rbOnExitTerminate"].set_active(True)
			self["vforce_dark_theme"].set_visible(True)
			self["lbl_vlanguage"].set_visible(True)
			self["vlanguage"].set_visible(True)
		# Check for filemanager python bindings current state of plugins
		status = []
		for widget_id in FM_DATA:
			so_file, package, plugin, location, name = FM_DATA[widget_id]
			if not get_fm_source_path(plugin) is None:
				if library_exists(so_file):
					self[widget_id].set_sensitive(True)
					self[widget_id].set_active(
						os.path.exists(get_fm_target_path(plugin, location))
					)
				else:
					log.warning("Cannot find %s required to support %s", so_file, name)
					status.append(_("Install %(package)s package to enable %(feature)s support") % {
						'package' : package,
						'feature' : name
					})
			else:
				log.warning("Cannot find %s.py required to support %s", plugin, name)
		self["fmLblIntegrationStatus"].set_text("\n".join(status))
		if StDownloader is None:
			for name in ("vst_autoupdate", "lblAutoupdate", "lblsyncthing_binary",
						 "lblsyncthing_binary2", "vsyncthing_binary", "btBrowse"):
				self[name].set_visible(False)
		self.cb_data_loaded(copy)
		self.cb_check_value()
	
	#@Overrides
	def display_value(self, key, w):
		if key == "vautostart_daemon":
			value = self.get_value(key[1:])
			if   value == 0: self["rbOnStartWait"].set_active(True)
			elif value == 1: self["rbOnStartRun"].set_active(True)
			else: self["rbOnStartAsk"].set_active(True)
		elif key == "vautokill_daemon":
			value = self.get_value(key[1:])
			if   value == 1: self["rbOnExitTerminate"].set_active(True)
			elif value == 0: self["rbOnExitLeave"].set_active(True)
			else: self["rbOnExitAsk"].set_active(True)
		else:
			return EditorDialog.display_value(self, key, w)
	
	#@Overrides
	def store_value(self, key, w):
		if key == "vautostart_daemon":
			if   self["rbOnStartWait"].get_active() : self.set_value(key[1:], 0)
			elif self["rbOnStartRun"].get_active() : self.set_value(key[1:], 1)
			else: return self.set_value(key[1:], 2)	# vOnStartAsk
		elif key == "vautokill_daemon":
			if self["rbOnExitTerminate"].get_active() : return self.set_value(key[1:], 1)
			elif self["rbOnExitLeave"].get_active() : return self.set_value(key[1:], 0)
			else: return self.set_value(key[1:], 2)	# vOnExitAsk
		elif key == "vst_autoupdate":
			# Reset updatecheck timer when autoupdate is turned on
			if self["vst_autoupdate"].get_active():
				self.values["last_updatecheck"] = LONG_AGO
			return EditorDialog.store_value(self, key, w)
		else:
			return EditorDialog.store_value(self, key, w)
	
	#@Overrides
	def set_value(self, key, value):
		if key == "autostart":
			set_run_on_startup(value, "Syncthing-GTK", get_executable(),
				"/usr/share/syncthing-gtk/icons/st-logo-128.png",
				"GUI for Syncthing")
		elif key == "daemon_priority":
			return EditorDialog.set_value(self, key, int(value))
		else:
			return EditorDialog.set_value(self, key, value)
	
	#@Overrides
	def get_value(self, key):
		if key == "autostart":
			return is_ran_on_startup("Syncthing-GTK")
		else:
			return EditorDialog.get_value(self, key)
	
	#@Overrides
	def on_data_loaded(self):
		self.values = self.config
		self.checks = {
			"vsyncthing_binary" : lambda p : os.path.isfile(p) and os.access(p, os.X_OK)
		}
		return self.display_values(VALUES)
	
	#@Overrides
	def update_special_widgets(self, *a):
		pass
	
	#@Overrides
	def on_save_requested(self):
		self.store_values(VALUES)
		# Save data to configuration file
		for k in self.values:
			self.app.config[k] = self.values[k]
		# Create / delete fm integration scripts
		for widget_id in FM_DATA:
			so_file, package, plugin, location, name = FM_DATA[widget_id]
			if self[widget_id].get_sensitive() and self[widget_id].get_active():
				# Should be enabled. Check if script is in place and create it if not
				source = get_fm_source_path(plugin)
				target = get_fm_target_path(plugin, location)
				if not source is None and not os.path.exists(target):
					try:
						# Create directory first
						os.makedirs(os.path.dirname(target))
					except Exception as e:
						# Ignore "file already exists" error
						pass
					try:
						if is_file_or_symlink(target):
							os.unlink(target)
						os.symlink(source, target)
						log.info("Created symlink '%s' -> '%s'", source, target)
					except Exception as e:
						log.error("Failed to symlink '%s' -> '%s'", source, target)
						log.error(e)
			else:
				# Should be disabled. Remove redundant scripts
				for extension in ("py", "pyc", "pyo"):
					target = get_fm_target_path(plugin, location, extension)
					if is_file_or_symlink(target):
						try:
							os.unlink(target)
							log.info("Removed '%s'", target)
						except Exception as e:
							log.error("Failed to remove '%s'", target)
							log.error(e)
		
		# Report work done
		self.syncthing_cb_post_config()
	
	#@Overrides
	def on_saved(self):
		self.close()
		# Recreate Notifications object if needed
		if HAS_DESKTOP_NOTIFY:
			if not self.app.notifications is None:
				self.app.notifications.kill()
				self.app.notifications = None
			if self.app.config["notification_for_update"] or self.app.config["notification_for_error"]:
				self.app.notifications = Notifications(self.app, self.app.daemon)
		# Restart or cancel updatecheck
		self.app.check_for_upgrade()
		# Update directory boxes to reflect change in 'folder_as_path'
		for rid in self.app.folders:
			box = self.app.folders[rid]
			title = box["path"] if self.app.config["folder_as_path"] else rid
			box.set_visible("id", self.app.config["folder_as_path"])
			box.set_title(title)

def library_exists(name):
	"""
	Checks if there is specified so file installed in one of known prefixes
	"""
	PREFIXES = [
		"/usr/lib64",	# Fedora
		"/usr/lib",
		"/usr/local/lib/",
		"/usr/x86_64-pc-linux-gnu/lib/",
		"/usr/i686-pc-linux-gnu/lib/",
		"/usr/lib/x86_64-linux-gnu/",
		"/usr/lib/i386-linux-gnu/",
	]
	for prefix in PREFIXES:
		if os.path.exists(os.path.join(prefix, name)):
			return True
	return False

def get_fm_target_path(plugin, location, extension="py"):
	"""
	Returns full path to plugin file in filemanager plugins directory
	"""
	datahome = os.path.expanduser("~/.local/share")
	if "XDG_DATA_HOME" in os.environ:
		datahome = os.environ["XDG_DATA_HOME"]
	return os.path.join(datahome, location, "%s.%s" % (plugin, extension))

def get_fm_source_path(plugin):
	"""
	Returns path to location where plugin file is installed
	"""
	filename = "%s.py" % (plugin,)
	paths = (
		# Relative path used while developing or when running
		# ST-GTK without installation
		"./scripts/",
		# Default installation path
		"/usr/share/syncthing-gtk",
		# Not-so default installation path
		"/usr/local/share/syncthing-gtk",
	)
	for path in paths:
		fn = os.path.abspath(os.path.join(path, filename))	
		if os.path.exists(fn):
			return fn
	return None

def is_file_or_symlink(path):
	"""
	Returns True if specified file exists, even as broken symlink.
	(os.path.exists() returns False for broken symlinks)
	"""
	if os.path.exists(path): return True
	try:
		os.readlink(path)
		return True
	except:
		pass
	return False

def browse_for_binary(parent_window, settings_dialog, value):
	"""
	Display file browser dialog to browse for syncthing binary.
	Used here and by FindDaemonDialog as well.
	"""
	# Prepare dialog
	d = Gtk.FileChooserDialog(
		_("Browse for Syncthing binary"),
		parent_window,
		Gtk.FileChooserAction.OPEN,
		(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
		Gtk.STOCK_OK, Gtk.ResponseType.OK))
	# Prepare filter
	f = Gtk.FileFilter()
	if IS_WINDOWS:
		f.set_name("Executables")
		f.add_pattern("*.exe")
	else:
		f.set_name("Binaries")
		f.add_mime_type("application/x-executable")
		f.add_mime_type("application/x-shellscript")
	d.add_filter(f)
	# Set default path
	confdir = os.path.join(get_config_dir(), "syncthing")
	prevvalue = str(settings_dialog[value].get_text()).strip()
	if prevvalue and os.path.exists(os.path.split(prevvalue)[0]):
		d.set_current_folder(os.path.split(prevvalue)[0])
	elif os.path.exists(confdir):
		d.set_current_folder(confdir)
	elif IS_WINDOWS:
		if "CommonProgramFiles" in os.environ:
			d.set_current_folder(os.environ["CommonProgramFiles"])
		elif os.path.exists("C:\\Program Files"):
			d.set_current_folder("C:\\Program Files")
		# Else nothing, just start whatever you like
	else:
		d.set_current_folder("/usr/bin")
	
	# Get response
	if d.run() == Gtk.ResponseType.OK:
		settings_dialog[value].set_text(d.get_filename())
	d.destroy()



#!/usr/bin/env python3
"""
Syncthing-GTK - Daemon

Class interfacing with syncthing daemon

Create instance, connect singal handlers and call daemon.reconnect()

"""


from gi.repository import Gio, GLib, GObject
from syncthing_gtk.timermanager import TimerManager
from syncthing_gtk.tools import parsetime, get_header, compare_version
from syncthing_gtk.tools import get_config_dir
from dateutil import tz
from xml.dom import minidom
from datetime import datetime
import json, os, sys, time, logging, urllib.parse
log = logging.getLogger("Daemon")

# Minimal version supported by Daemon class
MIN_VERSION = "0.14"

# Random constant used as key when adding headers to returned data in
# REST requests; Anything goes, as long as it isn't string
HTTP_HEADERS = int(65513)

# Last-seen values before this date are translated to never
NEVER = datetime(1971, 1, 1, 1, 1, 1, tzinfo=tz.tzlocal())

class Daemon(GObject.GObject, TimerManager):
	"""
	Object for interacting with syncthing daemon.
	
	List of signals:
		config-out-of-sync ()
			Emitted when daemon synchronization gets out of sync and
			daemon needs to be restarted.
		
		config-saved ()
			Emitted when daemon saves configuration without need for
			restarting.
		
		connected ()
			Emitted when connection to daemon is initiated, before
			configuration is loaded and parsed.
		
		disconnected (reason, message)
			Emitted after connection to daemon is lost. Connection can
			be reinitiated by calling reconnect()
				reason :	Daemon.SHUTDOWN if connection is closed
							after calling shutdown()
							Daemon.RESTART if connection is closed
							after calling restart()
							Daemon.UNEXPECTED for all other cases
				message:	generated error message
		
		config-loaded(config)
			Emitted while connection do daemon is being created, when
			configuration is loaded from daemon.
				config:		decoded /rest/config YAML file
		
		connection-error (reason, message, exception)
			Emitted if connection to daemon fails.
				reason:		Daemon.REFUSED if connection is refused and
							daemon probably offline. Connection will be
							retried automatically.
							Daemon.UNKNOWN for all other problems.
							Connection can be reinitiated by calling
							reconnect() in this case.
				message:	generated error message
				exception:	Exeception that caused problem or None
		
		my-id-changed (my_id, replaced)
			Emitted when ID is retrieved from device or when ID changes
			after client connects to another device
				my_id:		ID of device that is instance connected to.
		
		error (message)
			Emitted every time when daemon generates error readable by
			WebUI (/rest/errors call)
				message:	Error message sent by daemon
		
		folder-rejected(device_id, folder_id, label)
			Emitted when daemon detects unexpected folder from known
			device.
				device_id:	id of device that send unexpected folder id
				folder_id:	id of unexpected folder
				label:		label of unexpected folder or None
		
		device-rejected(device_id, device_name, address)
			Emitted when daemon detects connection from unknown device
				device_id:		device id
				device_name:	device name
				address:		address which connection come from
		
		device-added (id, name, used, data)
			Emited when new device is loaded from configuration
				id:		id of loaded device
				name:	name of loaded device (may be None)
				used:	true if there is any folder shared with this device
				data:	dict with rest of device data
		
		device-connected (id)
			Emitted when daemon connects to remote device
				id:			id of device
		
		device-disconnected (id)
			Emitted when daemon loses connection to remote device
				id:			id of device
		
		device-discovered (id, addresses)
			# TODO: What this event does?
				id:			id of device
				addresses:	list of device addresses
		
		device-data-changed (id, address, version, inbps, outbps, inbytes, outbytes)
			Emitted when device data changes
				id:			id of device
				address:	address of remote device
				version:	daemon version of remote device
				inbps:		download rate
				outbps:	upload rate
				inbytes:	total number of bytes downloaded
				outbytes:	total number of bytes uploaded
		
		last-seen-changed (id, last_seen)
			Emitted when daemon reported 'last seen' value for device changes
			or when is this value received for first time
				id:			id of device
				last_seen:	datetime object or None, if device was never seen
		
		device-paused (id):
			Emitted when synchronization with device is paused
				id:		id of folder
		
		device-resumed (id):
			Emitted when synchronization with device is resumed
				id:		id of folder
		
		device-sync-started (id, progress):
			Emitted after device synchronization is started
				id:			id of folder
				progress:	synchronization progress (0.0 to 1.0)
		
		device-sync-progress (id, progress):
			Emitted repeatedly while device is being synchronized
				id:			id of folder
				progress:	synchronization progress (0.0 to 1.0)
	
		device-sync-finished (id):
			Emitted after device synchronization is finished
				id:		id of folder
		
		folder-added (id, data)
			Emitted when new folder is loaded from configuration
				id:		id of loaded folder
				data:	dict with rest of folder data
		
		folder-error (id, errors)
			Emitted when when a folder cannot be successfully synchronized
				id:		id of loaded folder
				errors:	list with errors
		
		folder-data-changed (id, data):
			Emitted when change in folder data (/rest/model call)
			is detected and successfully loaded.
				id:		id of folder
				data:	dict with loaded data
		
		folder-data-failed (id):
			Emitted when daemon fails to load folder data (/rest/model call),
			most likely because folder was just added and syncthing
			daemon needs to be restarted
				id:		id of folder
		
		folder-scan-progress (id, progress):
			Emitted repeatedly while folder is being scanned
				id:			id of folder
				progress:	scan progress (0.0 to 1.0)
		
		folder-sync-progress (id, progress):
			Emitted repeatedly while folder is being synchronized
				id:			id of folder
				progress:	synchronization progress (0.0 to 1.0)
	
		folder-sync-finished (id):
			Emitted after folder synchronization is finished
				id:		id of folder
		
		folder-scan-started (id):
			Emitted after folder scan is started
				id:		id of folder
		
		folder-scan-finished (id):
			Emitted after folder scan is finished
				id:		id of folder
		
		folder-stopped (id, message):
			Emitted when folder enters 'stopped' state.
			No 'folder-sync', 'folder-sync-progress' and 'folder-scan-started'
			events are emitted after folder enters this state, until
			reconnect() is called.
				id:			id of folder
				message:	error message
		
		item-started (folder_id, filename, time):
			Emitted when synchronization of file starts
				folder_id:	id of folder that contains file
				filename:	synchronized file
				time:		event timestamp
		
		item-updated (folder_id, filename, time):
			Emited when change in local file is detected (LocalIndexUpdated event)
				folder_id:	id of folder that contains file
				filename:	updated file
				time:		event timestamp
		
		startup-complete():
			Emited when daemon initialization is complete.
		
		system-data-updated (ram_ussage, cpu_ussage, d_failed, d_total)
			Emitted every time when system information is received
			from daemon.
				ram_ussage:	memory ussage in bytes
				cpu_ussage:	CPU ussage in percent (0.0 to 100.0)
				d_failed:	Number of discovery servers that daemon failed to
							connect to
				d_total:	Total number of discovery servers
	"""
	
	
	__gsignals__ = {
		"config-out-of-sync"	: (GObject.SIGNAL_RUN_FIRST, None, ()),
		"config-saved"			: (GObject.SIGNAL_RUN_FIRST, None, ()),
		"connected"				: (GObject.SIGNAL_RUN_FIRST, None, ()),
		"disconnected"			: (GObject.SIGNAL_RUN_FIRST, None, (int, object)),
		"config-loaded"			: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"connection-error"		: (GObject.SIGNAL_RUN_FIRST, None, (int, object, object)),
		"error"					: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-rejected"		: (GObject.SIGNAL_RUN_FIRST, None, (object,object,object)),
		"device-rejected"		: (GObject.SIGNAL_RUN_FIRST, None, (object,object,object)),
		"my-id-changed"			: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"device-added"			: (GObject.SIGNAL_RUN_FIRST, None, (object, object, bool, object)),
		"device-connected"		: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"device-disconnected"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"device-discovered"		: (GObject.SIGNAL_RUN_FIRST, None, (object,object,)),
		"device-data-changed"	: (GObject.SIGNAL_RUN_FIRST, None, (object, object, object, float, float, object, object)),
		"last-seen-changed"		: (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
		"device-paused"			: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"device-resumed"		: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"device-sync-started"	: (GObject.SIGNAL_RUN_FIRST, None, (object, float)),
		"device-sync-progress"	: (GObject.SIGNAL_RUN_FIRST, None, (object, float)),
		"device-sync-finished"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-added"			: (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
		"folder-error"			: (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
		"folder-data-changed"	: (GObject.SIGNAL_RUN_FIRST, None, (object, object)),
		"folder-data-failed"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-sync-finished"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-sync-progress"	: (GObject.SIGNAL_RUN_FIRST, None, (object, float)),
		"folder-sync-started"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-scan-finished"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-scan-started"	: (GObject.SIGNAL_RUN_FIRST, None, (object,)),
		"folder-scan-progress"	: (GObject.SIGNAL_RUN_FIRST, None, (object, float)),
		"folder-stopped"		: (GObject.SIGNAL_RUN_FIRST, None, (object,object)),
		"item-started"			: (GObject.SIGNAL_RUN_FIRST, None, (object,object,object)),
		"item-updated"			: (GObject.SIGNAL_RUN_FIRST, None, (object,object,object)),
		"startup-complete"		: (GObject.SIGNAL_RUN_FIRST, None, ()),
		"system-data-updated"	: (GObject.SIGNAL_RUN_FIRST, None, (int, float, int, int)),
	}
	
	# Constants for 'reason' parameter of disconnected event
	UNEXPECTED	= 0 # connection closed by daemon
	SHUTDOWN	= 1
	RESTART		= 2
	
	# Constants for 'reason' parameter of connection-error event
	REFUSED			= 1
	NOT_AUTHORIZED	= 2
	OLD_VERSION		= 3
	TLS_UNSUPPORTED	= 4
	UNKNOWN			= 255
	
	def __init__(self, syncthing_configxml=None):
		GObject.GObject.__init__(self)
		TimerManager.__init__(self)
		self._CSRFtoken = None
		self._address = None
		self._api_key = None
		self._connected = False
		self._refresh_interval = 1 # seconds
		self._configxml = syncthing_configxml
		# syncing_folders holds set of folders that are being synchronized
		self._syncing_folders = set()
		# stopped_folders holds set of folders in 'stopped' state
		# No 'folder-sync', 'folder-sync-progress' and 'folder-scan-started'
		# events are emitted after folder enters this state
		self._stopped_folders = set()
		# syncing_devices does same thing, only for devices
		self._syncing_devices = set()
		# and once again, for folders in 'Scanning' state
		self._scanning_folders = set()
		# device_data stores data needed to compute transfer speeds
		# and synchronization state
		self._device_data = {}
		# folder_devices stores list of devices assigned to folder
		self._folder_devices = {}
		# last_seen holds last_seen value for each folder, preventing firing
		# last-seen-changed event with same values twice
		self._last_seen = {}
		# last_error_time is used to discard repeating errors
		self._last_error_time = None # Time is taken for first event
		# last_id is id of last event received from daemon
		self._last_id = 0
		# Epoch is increased when reconnect() method is called; It is
		# used to discard responses for old REST requests
		self._epoch = 1
		self._instance_id = None
		self._my_id = None
		self._read_config()
	
	### Internal stuff ###
	
	def _read_config(self):
		# Read syncthing config to get connection url
		if not self._configxml:
			self._configxml = os.path.join(get_config_dir(), "syncthing", "config.xml")
		if not os.path.exists(self._configxml) and os.path.exists(os.path.expanduser("~/snap/syncthing/common/syncthing")):
			# Special case for syncthing in snap package
			self._configxml = os.path.expanduser("~/snap/syncthing/common/syncthing/config.xml")
		try:
			log.debug("Reading syncthing config %s", self._configxml)
			with open(self._configxml, "r") as f:
				config = f.read()
		except Exception as e:
			raise InvalidConfigurationException("Failed to read daemon configuration: %s" % e)
		try:
			xml = minidom.parseString(config)
		except Exception as e:
			raise InvalidConfigurationException("Failed to parse daemon configuration: %s" % e)
		tls = "false"
		try:
			tls = xml.getElementsByTagName("configuration")[0] \
				.getElementsByTagName("gui")[0].getAttribute("tls")
		except Exception as e:
			pass
		self._tls = False
		self._cert = None
		if tls.lower() == "true":
			self._tls = True
			try:
				self._cert = Gio.TlsCertificate.new_from_file(
					os.path.join(get_config_dir(), "syncthing", "https-cert.pem"))
			except Exception as e:
				log.exception(e)
				raise TLSErrorException("Failed to load daemon certificate")
		try:
			self._address = xml.getElementsByTagName("configuration")[0] \
							.getElementsByTagName("gui")[0] \
							.getElementsByTagName("address")[0] \
							.firstChild.nodeValue
			if self._address.startswith("0.0.0.0"):
				addr, port = self._address.split(":", 1)
				self._address = "127.0.0.1:%s" % (port,)
				log.debug("WebUI listens on 0.0.0.0, connecting to 127.0.0.1 instead")
		except Exception as e:
			log.exception(e)
			raise InvalidConfigurationException("Required configuration node not found in daemon config file")
		try:
			self._api_key = xml.getElementsByTagName("configuration")[0] \
							.getElementsByTagName("gui")[0] \
							.getElementsByTagName("apikey")[0] \
							.firstChild.nodeValue
		except Exception as e:
			# API key can be none
			pass
	
	def override_config(self, address, api_key):
		"""
		Can be used to override settings loaded from config file.
		api_key can be None.
		"""
		self._address = address
		self.api_key = api_key
	
	def _get_device_data(self, nid):
		""" Returns dict with device data, creating it if needed """
		if not nid in self._device_data:
			self._device_data[nid] = {
					"inBytesTotal" : 0, "outBytesTotal" : 0,
					"inbps" : 0, "outbps" : 0 , "clientVersion" : "?",
					"address": "", "completion" : {}, "connected" : False,
				}
		return self._device_data[nid]
	
	def _request_config(self, *a):
		""" Request settings from syncthing daemon """
		RESTRequest(self, "system/config", self._syncthing_cb_config, self._syncthing_cb_config_error).start()
	
	def _request_folder_data(self, folder_id):
		id_enc = urllib.parse.quote(folder_id.encode('utf-8'))
		RESTRequest(self, "db/status?folder=%s" % (id_enc,), self._syncthing_cb_folder_data, self._syncthing_cb_folder_data_failed, folder_id).start()
	
	def _request_last_seen(self, *a):
		""" Request 'last seen' values for all devices """
		RESTRequest(self, "stats/device", self._syncthing_cb_last_seen).ignore_error().start()
	
	def _parse_dev_n_folders(self, config):
		"""
		Parses devices and folders from configuration and emits
		associated events.
		"""
		# Pre-parse folders to detect unused devices
		device_folders = {}
		for r in config["folders"]:
			rid = r["id"]
			for n in r["devices"]:
				nid = n["deviceID"]
				if not nid in device_folders : device_folders[nid] = []
				device_folders[nid].append(rid)

		# Parse devices
		for n in sorted(config["devices"], key=lambda x : x["name"].lower()):
			nid = n["deviceID"]
			self._get_device_data(nid)	# Creates dict with device data
			used = (nid in device_folders) and (len(device_folders[nid]) > 0)
			self.emit("device-added", nid, n["name"], used, n)
			
		# Parse folders
		for r in sorted(config["folders"], key=lambda x : x["id"].lower()):
			rid = r["id"]
			self._syncing_folders.add(rid)
			self._folder_devices[rid] = [ n["deviceID"] for n in r["devices"] ]
			self.emit("folder-added", rid, r)
			self._request_folder_data(rid)
	
	### Callbacks ###
	
	def _syncthing_cb_shutdown(self, data, reason):
		""" Callback for 'shutdown' AND 'restart' request """
		if 'ok' in data:
			if self._connected:
				self._connected = False
				self._epoch += 1
				self.emit("disconnected", reason, "")
			self.cancel_all()
	
	def _syncthing_cb_errors(self, errors):
		if errors["errors"] is not None:
			for e in errors["errors"]:
				if "time" in e:
					# TODO: Remove this next time support for older daemon is dropped
					t = parsetime(e["time"])
					msg = e["error"]
				elif "when" in e:
					t = parsetime(e["when"])
					msg = e["message"]
				else:
					# Can't decode this
					continue
				if t > self._last_error_time:
					self.emit("error", msg)
					self._last_error_time = t
		r = RESTRequest(self, "system/error", self._syncthing_cb_errors)
		self.timer("errors", self._refresh_interval * 5, r.start)
	
	def _syncthing_cb_connections(self, data, prev_time):
		now = time.time()
		td = now - prev_time
		
		cons = data["connections"]
		# Use my own device for totals, if it is already known
		# It it is not known, just skip totals for now
		if not self._my_id is None:
			cons[self._my_id] = data["total"]
		
		for id in cons:
			# Load device data
			nid = id
			device_data = self._get_device_data(nid)
			
			# Compute rates
			try:
				cons[id]["inbps"] = max(0.0, (cons[id]["inBytesTotal"] - device_data["inBytesTotal"]) / td);
				cons[id]["outbps"] = max(0.0, (cons[id]["outBytesTotal"] - device_data["outBytesTotal"]) / td);
			except Exception:
				cons[id]["inbps"] = 0.0
				cons[id]["outbps"] = 0.0
			# Store updated device_data
			for key in cons[id]:
				if not key in ('clientVersion', 'connected'):		# Don't want copy those
					if cons[id][key] != "":							# Happens for 'total'
						device_data[key] = cons[id][key]
			
			if "clientVersion" in cons[id] and cons[id]["clientVersion"] != "":
				device_data["clientVersion"] = cons[id]["clientVersion"]
			
			if cons[id]["paused"]:
				# Send "device-paused" signal if device needed
				device_data["connected"] = False
				self.emit("device-paused", nid)
			else:
				# Send "device-connected" signal, if device was disconnected until now
				if cons[id]["connected"]:
					if not device_data["connected"] and nid != self._my_id:
						device_data["connected"] = True
						self.emit("device-connected", nid)
			# Send "device-data-changed" signal
			self.emit("device-data-changed", nid, 
				device_data["address"],
				device_data["clientVersion"],
				device_data["inbps"],
				device_data["outbps"],
				device_data["inBytesTotal"],
				device_data["outBytesTotal"])
		
		# ... repeat until pronounced dead
		r = RESTRequest(self, "system/connections", self._syncthing_cb_connections, None, now)
		self.timer("conns", self._refresh_interval * 5, r.start)
	
	def _syncthing_cb_last_seen(self, data):
		for nid in data:
			if nid != HTTP_HEADERS:
				t = parsetime(data[nid]["lastSeen"])
				if t < NEVER: t = None
				if not nid in self._last_seen or self._last_seen[nid] != t:
					self._last_seen[nid] = t
					self.emit('last-seen-changed', nid, t)
	
	def _syncthing_cb_completion(self, data):
		nid = data["device"]
		rid = data["folder"]
		# Store acquired value
		device = self._get_device_data(nid)
		device["completion"][rid] = float(data["completion"])
		
		# Recompute stuff
		total = 100.0 * len(device["completion"])
		sync = 0.0
		if total > 0.0:
			sync = sum(device["completion"].values()) / total
		if sync <= 0 or sync >= 100:
			# Not syncing
			if nid in self._syncing_devices:
				self._syncing_devices.discard(nid)
				self.emit("device-sync-finished", nid)
		else:
			# Syncing
			if not nid in self._syncing_devices:
				self._syncing_devices.add(nid)
				self.emit("device-sync-started", nid, sync)
			else:
				self.emit("device-sync-progress", nid, sync)

	
	def _syncthing_cb_system(self, data):
		if "myID" not in data:
			# Invalid response
			r = RESTRequest(self, "system/status", self._syncthing_cb_system)
			log.warning("Invalid response received for rest/system/status request")
			self.timer("system", self._refresh_interval * 5, r.start)
			return
		
		if self._my_id != data["myID"]:
			if self._my_id != None:
				# Can myID be ever changed?
				log.warning("My ID has been changed on the fly")
			self._my_id = data["myID"]
			self.emit('my-id-changed', self._my_id)
			version = get_header(data[HTTP_HEADERS], "X-Syncthing-Version")
			if version:
				self._syncthing_cb_version_known(version)
			else:
				RESTRequest(self, "system/version", self._syncthing_cb_version).start()
		
		d_failed, d_total = 0, 0
		if "discoveryEnabled" in data and data["discoveryEnabled"]:
			d_total = data["discoveryMethods"]
			d_failed = len(data["discoveryErrors"])
		
		if "startTime" in data:
			if self._instance_id is None:
				self._instance_id = data["startTime"]
			else:
				if self._instance_id != data["startTime"]:
					return self._instance_replaced()
		
		self.emit('system-data-updated', data["sys"],
			float(data["cpuPercent"]), d_failed, d_total)
		
		r = RESTRequest(self, "system/status", self._syncthing_cb_system)
		self.timer("system", self._refresh_interval * 5, r.start)
	
	def _instance_replaced(self):
		"""
		Called when it is detected that syncthing daemon instance is no longer
		same as we talked to before
		"""
		log.warning("Daemon instance was replaced unexpectedly. Disconnecting from daemon.")
		self._disconnected(message="Daemon instance replaced unexpectedly")
	
	def _disconnected(self, reason=UNEXPECTED, message=""):
		""" Called to prepare and emit "disconnected" signal """
		self._my_id = None
		if self._connected:
			self._connected = False
			self._epoch += 1
			self.emit("disconnected", reason, message)
		self.cancel_all()
	
	def _syncthing_cb_version(self, data):
		if "version" in data:
			# New since https://github.com/syncthing/syncthing/commit/d7956dd4957fa6eee5971c072fd7181015fa876c
			version = data["version"]
		else:
			version = data["data"]
		self._syncthing_cb_version_known(version)
	
	def _syncthing_cb_version_known(self, version):
		"""
		Called when version is received from daemon, either by
		calling /rest/version or from X-Syncthing-Version header.
		"""
		if not compare_version(version, MIN_VERSION):
			# Syncting version too low. Cancel everything and report error
			self.cancel_all()
			self._epoch += 1
			msg = "daemon is too old"
			self.emit("connection-error", Daemon.OLD_VERSION, msg, Exception(msg))
			return
		if self._my_id != None:
			device = self._get_device_data(self._my_id)
			if version != device["clientVersion"]:
				device["clientVersion"] = version
				self.emit("device-data-changed", self._my_id, 
					None,
					device["clientVersion"],
					device["inbps"], device["outbps"],
					device["inBytesTotal"], device["outBytesTotal"])
	
	def _syncthing_cb_folder_data(self, data, rid):
		state = data['state']
		if state in ('error', 'stopped'):
			if not rid in self._stopped_folders:
				self._stopped_folders.add(rid)
				reason = data["invalid"] or data["error"]
				self.emit("folder-stopped", rid, reason)
		self.emit('folder-data-changed', rid, data)
		p = 0.0
		if state == "syncing":
			if float(data["globalBytes"]) > 0.0:
				p = float(data["inSyncBytes"]) / float(data["globalBytes"])
		self._folder_state_changed(rid, state, p)
	
	def _syncthing_cb_folder_data_failed(self, exception, request, rid):
		self.emit('folder-data-failed', rid)
	
	def _syncthing_cb_config(self, config):
		"""
		Called when configuration is loaded from syncthing daemon.
		After configuration is successfully parsed, app starts querying for events
		"""
		if not self._connected:
			self._connected = True
			self.emit('connected')
			
			self._parse_dev_n_folders(config)
			
			EventPollLoop(self).start()
			RESTRequest(self, "system/config/insync", self._syncthing_cb_config_in_sync).start()
			RESTRequest(self, "system/connections", self._syncthing_cb_connections, None, time.time()).start()
			RESTRequest(self, "system/status", self._syncthing_cb_system).start()
			self._request_last_seen()
			self.check_config()
			self.emit('config-loaded', config)
	
	def _syncthing_cb_config_error(self, exception, command):
		self.cancel_all()
		if isinstance(exception, GLib.GError):
			if exception.code in (0, 39, 34, 45):	# Connection Refused / Cannot connect to destination
				# It usually means that daemon is not yet fully started or not running at all.
				epoch = self._epoch
				self.emit("connection-error", Daemon.REFUSED, exception.message, exception)
				if epoch == self._epoch:
					r = RESTRequest(self, "system/config", self._syncthing_cb_config, self._syncthing_cb_config_error)
					self.timer("config", self._refresh_interval, r.start)
				return
		elif isinstance(exception, HTTPAuthException):
			self.emit("connection-error", Daemon.NOT_AUTHORIZED, exception.message, exception)
			return
		elif isinstance(exception, HTTPCode):
			# HTTP 404 may actually mean old daemon version
			version = get_header(exception.headers, "X-Syncthing-Version")
			if version != None and not compare_version(version, MIN_VERSION):
				self._epoch += 1
				msg = "daemon is too old"
				self.emit("connection-error", Daemon.OLD_VERSION, msg, Exception(msg))
			else:
				self.emit("connection-error", Daemon.UNKNOWN, exception.message, exception)
			return
		elif isinstance(exception, TLSUnsupportedException):
			self.emit("connection-error", Daemon.TLS_UNSUPPORTED, exception.message, exception)
			return
		elif isinstance(exception, ConnectionRestarted):
			# Happens on Windows. Just try again.
			GLib.idle_add(self._request_config)
			return
		elif isinstance(exception, TLSUnsupportedException):
			self.emit("connection-error", Daemon.TLS_UNSUPPORTED, exception.message, exception)
			return
		self.emit("connection-error", Daemon.UNKNOWN, exception.message, exception)
	
	def _syncthing_cb_config_in_sync(self, data):
		"""
		Handler for config/sync response. Emits 'config-out-of-sync' if
		configuration is not in sync.
		"""
		if "configInSync" in data:
			if not data["configInSync"]:
				# Not in sync...
				self.emit("config-out-of-sync")
	
	def _folder_state_changed(self, rid, state, progress):
		"""
		Emits event according to last known and new state.
		"""
		if state != "syncing" and rid in self._syncing_folders:
			self._syncing_folders.discard(rid)
			if not rid in self._stopped_folders:
				self.emit("folder-sync-finished", rid)
		if state != "scanning" and rid in self._scanning_folders:
			self._scanning_folders.discard(rid)
			if not rid in self._stopped_folders:
				self.emit("folder-scan-finished", rid)
		if state == "syncing":
			if not rid in self._stopped_folders:
				if rid in self._syncing_folders:
					self.emit("folder-sync-progress", rid, progress)
				else:
					self._syncing_folders.add(rid)
					self.emit("folder-sync-started", rid)
		elif state == "scanning":
			if not rid in self._stopped_folders:
				if not rid in self._scanning_folders:
					self._scanning_folders.add(rid)
					self.emit("folder-scan-started", rid)
	
	def _on_event(self, e):
		eType = e["type"]
		if eType in ("Ping", "Starting"):
			# Just ignore ignore those
			pass
		elif eType == "StartupComplete":
			self.emit("startup-complete")
		elif eType == "StateChanged":
			state = e["data"]["to"]
			rid = e["data"]["folder"]
			self._folder_state_changed(rid, state, 0)
		elif eType in ("RemoteIndexUpdated"):
			pass
		elif eType == "DeviceConnected":
			nid = e["data"]["id"]
			self.emit("device-connected", nid)
		elif eType == "DeviceDisconnected":
			   nid = e["data"]["id"]
			   self.emit("device-disconnected", nid)
		elif eType == "DeviceDiscovered":
			nid = e["data"]["device"]
			addresses = e["data"]["addrs"]
			self.emit("device-discovered", nid, addresses)
		elif eType == "DevicePaused":
			nid = e["data"]["device"]
			self.emit("device-paused", nid)
		elif eType == "DeviceResumed":
			nid = e["data"]["device"]
			self.emit("device-resumed", nid)
			self._request_last_seen()
		elif eType == "FolderRejected":
			nid = e["data"]["device"]
			rid = e["data"]["folder"]
			label = e["data"]["folderLabel"] if "folderLabel" in e["data"] else None
			self.emit("folder-rejected", nid, rid, label)
		elif eType == "DeviceRejected":
			nid = e["data"]["device"]
			name = e["data"]["name"]
			address = e["data"]["address"]
			self.emit("device-rejected", nid, name, address)
		elif eType == "FolderScanProgress":
			rid = e["data"]["folder"]
			total = float(e["data"]["total"])
			if total > 0:
				# ^^ just in case
				status = float(e["data"]["current"]) / total
				self.emit("folder-scan-progress", rid, status)
		elif eType == "ItemStarted":
			rid = e["data"]["folder"]
			filename = e["data"]["item"]
			t = parsetime(e["time"])
			self.emit("item-started", rid, filename, t)
		elif eType == "FolderCompletion":
			self._syncthing_cb_completion(e["data"])
		elif eType == "FolderSummary":
			rid = e["data"]["folder"]
			self._syncthing_cb_folder_data(e["data"]["summary"], rid)
		elif eType == "FolderErrors":
			rid = e["data"]["folder"]
			self.emit("folder-error", rid, e["data"]["errors"])
		elif eType == "ConfigSaved":
			self.emit("config-saved")
		elif eType == "ItemFinished":
			rid = e["data"]["folder"]
			if e["data"]["error"] is None:
				filename = e["data"]["item"]
				t = parsetime(e["time"])
				self.emit("item-updated", rid, filename, t)
		elif eType in ("ItemFinished", "DownloadProgress", "RelayStateChanged", "LocalIndexUpdated", "ListenAddressesChanged", "LoginAttempt"):
			# Not handled
			pass
		else:
			log.warning("Unhandled event type: %s", e)
	
	### External stuff ###
	
	def reconnect(self):
		"""
		Cancel all pending requests, throw away all data and (re)connect.
		Should be called from glib loop
		"""
		self.close()
		GLib.idle_add(self._request_config)
	
	def reload_config(self, callback=None, error_callback=None):
		"""
		Reloads config from syncthing daemon.
		Calling this will cause or may cause emiting following events
		with reloaded data:
		- folder-added
		- device-added
		- config-out-of-sync
		"""
		def reload_config_cb(config):
			self._parse_dev_n_folders(config)
			if not callback is None:
				callback()
			RESTRequest(self, "system/config/insync", self._syncthing_cb_config_in_sync).start()
		RESTRequest(self, "system/config", reload_config_cb, error_callback).start()
	
	def close(self):
		"""
		Terminates everything, cancel all pending requests, throws away
		data.
		Works like reconnect(), but without reconnecting.
		"""
		self._my_id = None
		self._instance_id = None
		self._connected = False
		self._syncing_folders = set()
		self._stopped_folders = set()
		self._syncing_devices = set()
		self._scanning_folders = set()
		self._device_data = {}
		self._folder_devices = {}
		self._last_id = 0
		self._last_seen = {}
		self.cancel_all()
		self._epoch += 1
	
	def check_config(self):
		"""
		Check if configuration is in sync.
		Should cause 'config-out-of-sync' event to be raised ASAP.
		"""
		RESTRequest(self, "system/config/insync", self._syncthing_cb_config_in_sync).start()
	
	def read_config(self, callback, error_callback=None, *calbackdata):
		"""
		Asynchronously reads last configuration version from daemon
		(even if this version is not currently used). Calls
		callback(config) with data decoded from json on success,
		error_callback(exception) on failure
		"""
		RESTRequest(self, "system/config", callback, error_callback, *calbackdata).start()
	
	def write_config(self, config, callback, error_callback=None, *calbackdata):
		"""
		Asynchronously POSTs new configuration to daemon. Calls
		callback() on success, error_callback(exception) on failure.
		Should cause 'config-out-of-sync' event to be raised ASAP.
		"""
		def run_before(data, *a):
			self.check_config()
			callback(*calbackdata)
		RESTPOSTRequest(self, "system/config", config, run_before, error_callback, *calbackdata).start()
	
	def read_stignore(self, folder_id, callback, error_callback=None, *calbackdata):
		"""
		Asynchronously reads .stignore data from from daemon.
		Calls callback(text) with .stignore content on success,
		error_callback(exception) on failure
		"""
		def r_filter(data, *a):
			if "ignore" in data and not data["ignore"] is None:
				callback("\n".join(data["ignore"]).strip(" \t\n"), *a)
			else:
				callback("", *a)
		id_enc = urllib.parse.quote(folder_id.encode('utf-8'))
		RESTRequest(self, "db/ignores?folder=%s" % (id_enc,), r_filter, error_callback, *calbackdata).start()
	
	def write_stignore(self, folder_id, text, callback, error_callback=None, *calbackdata):
		"""
		Asynchronously POSTs .stignore to daemon. Calls callback()
		with on success, error_callback(exception) on failure.
		"""
		data = { 'ignore': text.split("\n") }
		id_enc = urllib.parse.quote(folder_id.encode('utf-8'))
		RESTPOSTRequest(self, "db/ignores?folder=%s" % (id_enc,), data, callback, error_callback, *calbackdata).start()
	
	def restart(self):
		"""
		Asks daemon to restart. If successful, call will cause
		'disconnected' event with Daemon.RESTART reason to be fired
		"""
		RESTPOSTRequest(self, "system/restart",  {}, self._syncthing_cb_shutdown, None, Daemon.RESTART).start()
	
	def shutdown(self):
		"""
		Asks daemon to shutdown. If successful, call will cause
		'disconnected' event with Daemon.SHUTDOWN reason to be fired
		"""
		RESTPOSTRequest(self, "system/shutdown",  {}, self._syncthing_cb_shutdown, None, Daemon.SHUTDOWN).start()
	
	def syncing(self):
		""" Returns true if any folder is being synchronized right now  """
		return len(self._syncing_folders) > 0
	
	def get_api_key(self):
		""" Returns API key used for communication with daemon. May return None """
		return self._api_key
	
	def get_min_version(self):
		"""
		Returns minimal syncthing daemon version that daemon instance
		can handle.
		"""
		return MIN_VERSION
	
	def get_syncing_list(self):
		"""
		Returns list of ids of foldersitories that are being
		synchronized right now.
		"""
		return list(self._syncing_folders)
	
	def get_my_id(self):
		"""
		Returns ID of device that is instance connected to.
		May return None to indicate that ID is not yet known
		"""
		return self._my_id
	
	def get_version(self):
		"""
		Returns daemon version or "unknown" if daemon version is not yet
		known
		"""
		if self._my_id == None: return "unknown"
		device = self._get_device_data(self._my_id)
		if "clientVersion" in device:
			return device["clientVersion"]
		return "unknown"
	
	def get_webui_url(self):
		""" Returns web ui url in http(s)://127.0.0.1:8080 format """
		return "%s://%s" % (
			"https" if self._tls else "http",
			self._address
		)
	
	def get_address(self):
		""" Returns tuple address on which daemon listens on. """
		return self._address
	
	def is_connected(self):
		""" Returns True if daemon is known to be alive """
		return self._connected
	
	def pause(self, device_id):
		""" Pauses synchronization with specified device """
		RESTPOSTRequest(self, "system/pause?device=%s" % (device_id,), {}, lambda *a: a, lambda *a: log.error(a), device_id).start()
	
	def resume(self, device_id):
		""" Resumes synchronization with specified device """
		RESTPOSTRequest(self, "system/resume?device=%s" % (device_id,), {}, lambda *a: a, lambda *a: log.error(a), device_id).start()
	
	def rescan(self, folder_id, path=None):
		""" Asks daemon to rescan entire folder or specified path """
		if path is None:
			id_enc = urllib.parse.quote(folder_id.encode('utf-8'))
			RESTPOSTRequest(self, "db/scan?folder=%s" % (id_enc,), {}, lambda *a: a, lambda *a: log.error(a), folder_id).start()
		else:
			url = "db/scan?folder=%s&sub=%s" % (
				urllib.parse.quote(folder_id.encode('utf-8')),
				urllib.parse.quote(path.encode('utf-8'))
			)
			RESTPOSTRequest(self, url, {}, lambda *a: a, lambda *a: log.error(a), folder_id).start()
	
	def override(self, folder_id):
		""" Asks daemon to override remote changes made in specified folder """
		id_enc = urllib.parse.quote(folder_id.encode('utf-8'))
		RESTPOSTRequest(self, "db/override?folder=%s" % (id_enc,), {}, lambda *a: a, lambda *a: log.error(a), folder_id).start()
	
	def revert(self, folder_id):
		""" Asks daemon to revert local changes made in specified folder """
		id_enc = urllib.quote(folder_id.encode('utf-8'))
		RESTPOSTRequest(self, "db/revert?folder=%s" % (id_enc,), {}, lambda *a: a, lambda *a: log.error(a), folder_id).start()
	
	def request_events(self):
		"""
		No longer needed.
		"""
		pass
	
	def set_refresh_interval(self, i):
		""" Sets interval used mainly by event querying timer """
		self._refresh_interval = i
		log.verbose("Set refresh interval to %s", i)


class RESTRequest(Gio.SocketClient):
	"""
	REST over HTTP(s) request. Handles everything and calls callback with response
	received. It is assumed that response will always be JSON-encoded and
	it is automatically decoded.
	
	If request fails and error_callback is not set, it is automatically repeated.
	
	Callback signatures:
		callback(json_data, callback_data... )
		error_callback(exception, command, callback_data... )
	"""
	
	def __init__(self, parent, command, callback, error_callback=None, *callback_data):
		Gio.SocketClient.__init__(self, tls=parent._tls)
		self._callback = callback
		self._error_callback = error_callback
		self._command = command
		self._parent = parent
		self._connection = None
		self._callback_data = callback_data or ()
		if parent._tls:
			GObject.Object.connect(self, "event", self._socket_event)
	
	def _connected(self, _self, results):
		""" Called after TCP connection is initiated """
		try:
			self._connection = self.connect_to_service_finish(results)
			if self._connection == None:
				raise Exception("Unknown error")
		except Exception as e:
			log.exception(e)
			if hasattr(e, "domain") and e.domain == "g-tls-error-quark":
				e = TLSUnsupportedException(e.message)
			self._error(e)
			return
		if self._epoch != self._parent._epoch:
			# Too late, throw it away
			self._connection.close(None)
			log.verbose("Discarded old connection for %s", self._command)
			return
		if self._parent._CSRFtoken is None and self._parent._api_key is None:
			# Request CSRF token first
			log.verbose("Requesting cookie")
			get_str = "\r\n".join([
				"GET / HTTP/1.0",
				"Host: %s" % self._parent._address,
				(("X-API-Key: %s" % self._parent._api_key) if not self._parent._api_key is None else "X-nothing: x"),
				"Connection: close",
				"",
				"",
				]).encode("utf-8")
		else:
			self._send_request()
	
	def _send_request(self):
		get_str = self._format_request()
		try:
			self._connection.get_output_stream().write_all(get_str, None)
		except Exception as e:
			self._error(e)
			return
		self._connection.get_input_stream().read_bytes_async(102400, 1, None, self._response)
	
	def _parse_csrf(self, response):
		for d in response:
			if d.startswith(b"Set-Cookie:"):
				for c in d.split(b":", 1)[1].split(b";"):
					if c.strip().startswith(b"CSRF-Token-"):
						self._CSRFtoken = c.strip(b" \r\n").decode()
						log.verbose("Got new cookie: %s", self._CSRFtoken)
						break
				if self._CSRFtoken != None:
					break
	
	def _format_request(self):
		"""
		Formats HTTP request (GET /xyz HTTP/1.0... ) before sending it to daemon
		"""
		return "\r\n".join([
			"GET /rest/%s HTTP/1.0" % self._command,
			"Host: %s" % self._parent._address,
			"Cookie: %s" % self._parent._CSRFtoken,
			(("X-%s" % self._parent._CSRFtoken.replace("=", ": ")) if self._parent._CSRFtoken else "X-nothing: x"),
			(("X-API-Key: %s" % self._parent._api_key) if not self._parent._api_key is None else "X-nothing2: x"),
			"Connection: close",
			"", ""
			]).encode("utf-8")
	
	def _response(self, stream, results):
		try:
			response = stream.read_bytes_finish(results)
			if response == None:
				raise Exception("No data received")
		except Exception as e:
			self._connection.close(None)
			self._error(e)
			return
		if self._epoch != self._parent._epoch:
			# Too late, throw it away
			self._connection.close(None)
			log.verbose("Discarded old response for %s", self._command)
			return
		# Repeat read_bytes_async until entire response is read into buffer
		self._buffer.append(response.get_data())
		if response.get_size() > 0:
			self._connection.get_input_stream().read_bytes_async(102400, 1, None, self._response)
			return
		self._connection.close(None)
		response, self._buffer = (b"".join(self._buffer)), []
		if self._parent._CSRFtoken is None and self._parent._api_key is None:
			# I wanna cookie!
			self._parse_csrf(response.split(b"\n"))
			if self._parent._CSRFtoken == None:
				# This is pretty fatal and likely to fail again,
				# so request is not repeated automatically
				if self._error_callback == None:
					log.error("Request '%s' failed: Error: failed to get CSRF cookie from daemon", self._command)
				else:
					self._error(Exception("Failed to get CSRF cookie"))
				return
			# Repeat request with acquired cookie
			self.start()
			return
		# Split headers from response
		headers, response = self._split_headers(response)
		if headers is None: return
		# Parse response and call callback
		try:
			rdata = json.loads(response)
		except IndexError: # No data
			rdata = { }
		except ValueError: # Not a JSON
			rdata = {'data' : response.decode('utf-8', errors='ignore') }
		if type(rdata) == dict:
			rdata[HTTP_HEADERS] = headers
		self._callback(rdata, *self._callback_data)
	
	def _split_headers(self, buffer):
		try:
			headers, response = buffer.split(b"\r\n\r\n", 1)
			headers = headers.decode('utf-8', errors='ignore').split("\r\n")
			code = int(headers[0].split(" ")[1])
			if code == 401:
				self._error(HTTPAuthException(buffer))
				return None, None
			elif code == 404:
				self._error(HTTPCode(404, "Not found", buffer, headers))
				return None, None
			elif code != 200:
				self._error(HTTPCode(code, response, buffer, headers))
				return None, None
		except Exception as e:
			# That probably wasn't HTTP
			import traceback
			traceback.print_exc()
			self._error(InvalidHTTPResponse(buffer))
			return None, None
		return headers, response
	
	def _error(self, exception):
		""" Error handler for _response method """
		if self._connection:
			self._connection.close(None)
			self._connection = None
		if self._error_callback:
			if self._epoch != self._parent._epoch:
				exception = ConnectionRestarted()
			self._error_callback(exception, self._command, *self._callback_data)
		elif self._epoch == self._parent._epoch:
			try:
				log.error("Request '%s' failed (%s); Repeating...", self._command, exception)
			except UnicodeDecodeError:
				# Windows...
				log.error("Request '%s' failed; Repeating...", self._command)
			self._parent.timer(str(self), 1, self.start)
	
	def _socket_event(self, _self, event, connectable, con):
		""" Setups TSL certificate if HTTPS is used """
		if event == Gio.SocketClientEvent.TLS_HANDSHAKING:
			con.connect("accept-certificate", self._accept_certificate)
	
	def _accept_certificate(self, con, peer_cert, errors):
		""" Check if server presents expected certificate and accept connection """
		return peer_cert.is_same(self._parent._cert)
	
	def ignore_error(self):
		"""
		Causes RESTRequest to ignore any error - no callback is called and
		request is *not* autorepeated.
		
		Returns self.
		"""
		self._error_callback = lambda *a: True
		return self
	
	def start(self):
		if self._parent._address.startswith("127.0.0.1"):
			self.set_enable_proxy(False)
		self._epoch = self._parent._epoch
		self._buffer = []
		self.connect_to_host_async(self._parent._address, 0, None, self._connected)
		return self


class RESTPOSTRequest(RESTRequest):
	""" Similar to RESTRequest, but this one uses HTTP POST and sends data """
	def __init__(self, parent, command, data, callback, error_callback=None, *callback_data):
		RESTRequest.__init__(self, parent, command, callback, error_callback, *callback_data)
		self._data = data
	
	def _format_request(self):
		"""
		Formats POST request before sending it to daemon
		"""
		json_str = json.dumps(self._data)
		return "\r\n".join([
			"POST /rest/%s HTTP/1.0" % self._command,
			"Host: %s" % self._parent._address,
			"Cookie: %s" % self._parent._CSRFtoken,
			(("X-%s" % self._parent._CSRFtoken.replace("=", ": ")) if self._parent._CSRFtoken else "X-nothing: x"),
			(("X-API-Key: %s" % self._parent._api_key) if not self._parent._api_key is None else "X-nothing2: x"),
			"Content-Length: %s" % len(json_str),
			"Content-Type: application/json",
			"Connection: close",
			"",
			json_str
			]).encode("utf-8")


class EventPollLoop(RESTRequest):
	"""
	Event polling 'loop' continuously polls events from daemon using one HTTP(s)
	connection. If connection is broken, EventPollLoop reconnects automatically.
	'Loop' is canceled automatically when parent _epoch is increased.
	"""
	def __init__(self, parent):
		RESTRequest.__init__(self, parent, "events", None, None)
		self._last_event_id = -1
	
	def _format_request(self):
		"""
		Event request is as special as it gets, with HTTP/1.1, connection held
		and continuously requesting more and more data.
		"""
		if self._last_event_id < 0:
			url = "/rest/events?limit=1"
		else:
			url = "/rest/events?since=%s" % (self._last_event_id,)
		
		return "\r\n".join([
			"GET %s HTTP/1.1" % url,
			"Host: %s" % self._parent._address,
			"Cookie: %s" % self._parent._CSRFtoken,
			(("X-%s" % self._parent._CSRFtoken.replace("=", ": ")) if self._parent._CSRFtoken else "X-nothing: x"),
			(("X-API-Key: %s" % self._parent._api_key) if not self._parent._api_key is None else "X-nothing2: x"),
			"Cache-Control: no-cache",
			"Connection: keep-alive",
			"Pragma: no-cache",
			"", ""
			]).encode("utf-8")
	
	def _error(self, exception):
		if self._connection:
			self._connection.close(None)
		if self._epoch == self._parent._epoch:
			if isinstance(exception, GLib.GError):
				if exception.code in (0, 39, 34):	# Connection terminated unexpectedly, Connection Refused
					self._parent._disconnected(message=str(exception))
					return
			self._parent.timer(None, 1, self.start)
	
	def start(self):
		RESTRequest.start(self)
		self._buffer = b""
		
	def _response(self, stream, results):
		if self._parent._CSRFtoken is None and self._parent._api_key is None:
			return RESTRequest._response(self, stream, results)
		try:
			response = stream.read_bytes_finish(results)
			if response == None:
				raise Exception("No data received")
		except Exception as e:
			return self._error(e)
		if self._epoch != self._parent._epoch:
			self._connection.close(None)
			return
		
		buffer = response.get_data()
		assert type(buffer) == bytes
		headers, response = self._split_headers(buffer)
		if headers is None: return
		headers = { x : y.strip() for (x,y) in [ h.split(":", 1) for h in headers if ":" in h ] }
		if "Transfer-Encoding" not in headers or headers["Transfer-Encoding"] != "chunked":
			# Something just went horribly wrong
			self._error(InvalidHTTPResponse(buffer))
			return
		self._buffer = response
		self._chunk_size = -1
		self._parse_chunk()
	
	def _chunk(self, stream, results):
		try:
			response = stream.read_bytes_finish(results)
			if response == None:
				raise Exception("nothing")
		except Exception as e:
			return self._error(e)
		if self._epoch != self._parent._epoch:
			self._connection.close(None)
			return
		data = response.get_data()
		if len(data) == 0:
			# Connection broken
			self._connection.close(None)
			return self.start()
		self._buffer += data
		self._parse_chunk()
	
	def _resend_request(self):
		""" Sends another request using same connection """
		self._chunk_size = -1
		get_str = "\r\n".join([
			"GET /rest/events?since=%s HTTP/1.1" % (self._last_event_id,),
			"Host: %s" % self._parent._address,
			(("X-API-Key: %s" % self._parent._api_key) if not self._parent._api_key is None else "X-nothing: x"),
			"",
			"",
			]).encode("utf-8")
		try:
			self._connection.get_output_stream().write_all(get_str, None)
		except Exception as e:
			self._connection.close(None)
			return self.start()
		self._connection.get_input_stream().read_bytes_async(10, 1, None, self._chunk)
	
	
	def _parse_chunk(self):
		if self._chunk_size < 0:
			try:
				# Try to decode chunk. May raise exception if only very few bytes
				# have been read so far
				size_str, rest = self._buffer.split(b"\r\n", 1)
				self._chunk_size = int(size_str, 16)
				self._buffer = rest
				if self._chunk_size < 1:
					# Zero-sized chunk means end of transfer
					return self._resend_request()
				self._chunk_size += 2	# 2b for following \r\n
			except (ValueError, IndexError):
				self._chunk_size = -1
				self._connection.get_input_stream().read_bytes_async(100, 1, None, self._chunk)
				return
		retrieved = len(self._buffer)
		if retrieved < self._chunk_size:
			self._connection.get_input_stream().read_bytes_async(self._chunk_size - retrieved, 1, None, self._chunk)
			return
		
		response, self._buffer = self._buffer[0:retrieved], self._buffer[retrieved:]
		try:
			events = json.loads(response)
		except Exception:
			# Invalid response
			self._connection.close(None)
			return self.start()
		
		for event in events:
			if self._last_event_id >= 0 and event["id"] != self._last_event_id + 1:
				# Event IDs are not continuous, something just went horribly wrong
				# There is only one case when this is expected: When connection
				# to daemon is lost and ST-GTK unknowingly reconnects to
				# different instance.
				return self._parent._instance_replaced()
			self._last_event_id = event["id"]
			self._parent._on_event(event)
			sys.stdout.flush()
		
		self._resend_request()



class InvalidConfigurationException(RuntimeError): pass
class TLSUnsupportedException(RuntimeError): pass
class TLSErrorException(RuntimeError): pass

class HTTPError(RuntimeError):
	def __init__(self, message, full_response):
		RuntimeError.__init__(self, message)
		self.full_response = full_response

class InvalidHTTPResponse(HTTPError):
	def __init__(self, full_response):
		HTTPError.__init__(self, "Invalid HTTP response", full_response)

class HTTPCode(HTTPError):
	def __init__(self, code, message, full_response, headers = []):
		HTTPError.__init__(self, "HTTP error %s : %s" % (code, message), full_response)
		self.code = code
		self.message = message
		self.headers = headers
	def __str__(self):
		if self.message is None:
			return "HTTP/%s" % (self.code,)
		else:
			return "HTTP/%s: %s" % (self.code, self.message)

class HTTPAuthException(HTTPCode):
	def __init__(self, full_response):
		HTTPCode.__init__(self, 401, None, full_response)
	def __str__(self):
		return "HTTP/401 Unauthorized"

class ConnectionRestarted(Exception):
	def __init__(self):
		Exception.__init__(self, "Connection was restarted after request")


if __name__ == "__main__":
	# Small thing for testing
	from .tools import init_logging, set_logging_level
	init_logging()
	set_logging_level(True, True)
	daemon = Daemon()
	daemon.connect('connected', lambda *a: sys.stdout.write("*** connected ***\n"))
	daemon.connect('disconnected', lambda *a: sys.stdout.write("*** disconnected ***\n"))
	daemon.reconnect()
	GLib.MainLoop().run()



#!/usr/bin/env python2
"""
Caja plugin for Syncthing-GTK
See syncthing_gtk/nautilusplugin.py for more info
"""

from gi.repository import Caja

# Setting this environment variable will prevent __init__ in
# syncthing_gtk package from loading stuff that depends on GTK3-only
# features. It probably breaks other modules in most horrible ways,
# but they are not going to be used anyway
import os
os.environ["GTK2APP"] = "1"

from syncthing_gtk.nautilusplugin import NautiluslikeExtension

NautiluslikeExtension.set_plugin_module(Caja)
class CajaNautilus(NautiluslikeExtension, Caja.InfoProvider, Caja.MenuProvider):
	pass



#!/usr/bin/env python3
"""
Syncthing-GTK - 1st run wizard

Basically runs Syncthing daemon with -generate option and setups some
values afterwards.
"""


from gi.repository import Gtk, GLib, GdkPixbuf
from syncthing_gtk.daemonoutputdialog import DaemonOutputDialog
from syncthing_gtk.daemonprocess import DaemonProcess
from syncthing_gtk.configuration import Configuration
from syncthing_gtk.tools import get_config_dir, IS_WINDOWS, is_portable
from syncthing_gtk.tools import can_upgrade_binary, compare_version
from syncthing_gtk.tools import _ # gettext function
try:
	from syncthing_gtk.stdownloader import StDownloader
except ImportError:
	StDownloader = None

import os, socket, random, string, bcrypt
import logging, traceback, platform
from xml.dom import minidom

log = logging.getLogger("Wizard")

DEFAULT_PORT = 8080
MAX_PORT = 8100

class Wizard(Gtk.Assistant):
	def __init__(self, gladepath="/usr/share/syncthing-gtk",
				iconpath="/usr/share/syncthing-gtk/icons", config=None):
		# Init
		Gtk.Assistant.__init__(self)
		if not config is None:
			self.config = config
		else:
			self.config = Configuration()
		self.gladepath = gladepath
		self.iconpath = iconpath
		self.syncthing_options = {}
		self.lines = []					# Daemon and wizard output,
										# maybe for error reports
		self.finished = False
		self.connect("prepare", self.prepare_page)
		# Find syncthing configuration directory
		self.st_configdir = os.path.join(get_config_dir(), "syncthing")
		self.st_configfile = os.path.join(get_config_dir(), "syncthing", "config.xml")
		# Window setup
		self.set_position(Gtk.WindowPosition.CENTER)
		self.set_size_request(720, -1)
		self.set_default_size(720, 300)
		self.set_deletable(True)
		if IS_WINDOWS:
			self.set_icon_list([GdkPixbuf.Pixbuf.new_from_file("icons/32x32/apps/syncthing-gtk.png")])
		else:
			self.set_icon_name("syncthing-gtk")
		self.set_title("%s %s" % (_("Syncthing-GTK"), _("First run wizard")))
		# Add "Quit" button
		self.quit_button = Gtk.Button.new_from_stock("gtk-quit")
		self.add_action_widget(self.quit_button)
		self.quit_button.set_visible(True)
		self.quit_button.connect("clicked", lambda *a : self.emit("cancel"))
		# Pages
		self.add_page(IntroPage(self))
		self.add_page(FindDaemonPage())
		self.add_page(GenerateKeysPage())
		self.add_page(HttpSettingsPage())
		self.add_page(SaveSettingsPage())
		self.add_page(LastPage())
	
	def add_page(self, page):
		""" Adds page derived from custom Page class """
		index = self.append_page(page)
		page.parent = self
		self.set_page_type(page, page.TYPE)
		self.set_page_title(page, _(page.TITLE) + "  ")
		return index
	
	def insert(self, page):
		"""
		Inserts new page after currently displayed.
		"""
		index = self.get_current_page()
		index = self.insert_page(page, index + 1)
		page.parent = self
		self.set_page_type(page, page.TYPE)
		self.set_page_title(page, _(page.TITLE) + "  ")
		return index
	
	def insert_and_go(self, page):
		"""
		Inserts new page after currently displayed
		and switches to it.
		"""
		index = self.insert(page)
		self.set_current_page(index)
		return index
	
	def only_page(self, page):
		for r in range(1, page):
			self.remove_page(r)
		while self.get_n_pages() > 3:
			self.remove_page(self.get_n_pages() - 2)
		self.set_current_page(1)
	
	def prepare_page(self, another_self, page):
		""" Called before page is displayed """
		self.commit() # Prevents back button from being displayed
		page.prepare()
	
	def find_widget(self, compare_fn, parent=None):
		"""
		Recursively searches for widget, returning first one
		for which compare_fn(widget) returns True
		"""
		if parent is None : parent = self
		for w in parent.get_children():
			if compare_fn(w): return w
			if isinstance(w, Gtk.Container):
				r = self.find_widget(compare_fn, w)
				if not r is None: return r
		return None
	
	def output_line(self, line):
		""" Called for every line that wizard or daemon process outputs """
		self.lines.append(line)
		log.info(line)
	
	def error(self, page, title, message, display_bugreport_link):
		"""
		Called from pages on error. Removes everything from page and
		creates error message.
		"""
		for c in [] + page.get_children() :
			page.remove(c)
		# Title
		l_title = WrappedLabel("<b>%s</b>" % (title,))
		l_title.props.margin_bottom = 15
		page.attach(l_title,	0, 0, 2, 1)
		# Message
		l_message = WrappedLabel(message)
		l_message.props.margin_bottom = 15
		page.attach(l_message,	0, 1, 2, 1)
		# Bugreport link
		if display_bugreport_link:
			github_link = '<a href="https://github.com/syncthing/syncthing-gtk/issues">GitHub</a>'
			l_bugreport = WrappedLabel(
				_("Please, check error log and fill bug report on %s.") % (github_link,)
			)
			page.attach(l_bugreport, 0, 2, 2, 1)
			# 'Display error log' button
			button = Gtk.Button(_("Display error log"))
			button.props.margin_top = 25
			page.attach(button, 1, 3, 2, 1)
			button.connect("clicked", lambda *a : self.show_output())
		
		page.show_all()
		return page
	
	def show_output(self, *a):
		"""
		Displays DaemonOutput window with error messages captured
		during key generation.
		"""
		d = DaemonOutputDialog(self, None)
		d.show_with_lines(self.lines, self)
	
	def is_finished(self):
		""" Returns True if user finished entire wizard """
		return self.finished
	
	def run(self, *a):
		self.show()
		self.connect('cancel', Gtk.main_quit)
		self.connect('close', Gtk.main_quit)
		Gtk.main()

class WrappedLabel(Gtk.Label):
	def __init__(self, markup):
		Gtk.Label.__init__(self)
		self.set_justify(Gtk.Justification.LEFT)
		self.set_line_wrap(True)
		self.set_markup(markup)
		self.set_alignment(0, 0.5)

# @AbstractClass
class Page(Gtk.Grid):
	# TYPE = <needs to be defined in derived class>
	# TITLE = <needs to be defined in derived class>
	def __init__(self, dialog=None):
		Gtk.Grid.__init__(self)
		self.dialog = dialog
		self.parent = None
		self.init_page()
		self.show_all()
	
	def prepare(self):
		""" Sets page as complete by default """
		self.parent.set_page_complete(self, True)

	def init_page(self):
		""" Initialize the page """
		raise NotImplementedError()

class IntroPage(Page):
	TYPE = Gtk.AssistantPageType.INTRO
	TITLE = "Intro"
	def init_page(self):
		""" First, intro page. Just static text that explains what's going on """ 
		config_folder = "~/.config/syncthing"
		config_folder_link = '<a href="file://%s">%s</a>' % (
				os.path.expanduser(config_folder), config_folder)
		next_label = _("Next")
		try:
			# Hacky way to determine label on 'Next' button
			all_buttons = [ b for b in self.dialog.quit_button.get_parent().get_children()
				if isinstance(b, Gtk.Button) and b is not self.dialog.quit_button ]
			# order is 'apply, next, back, finish, cancel'
			next_label = all_buttons[1].get_label().replace("_", "")
		except: pass
		# TODO: Temporal solution, remove later
		changed_line = _("Continue this wizard to create a Syncthing configuration file or abort it to exit.")
		if "%s" in changed_line:
			changed_line = changed_line % (next_label,)
		
		self.attach(WrappedLabel(
			"<b>" + _("Welcome to Syncthing-GTK first run wizard!") + "</b>" +
			"\n\n" +
			_("It looks like you never have used Syncthing.") + " " +
			_("Initial configuration should be created.") +  " " +
			changed_line + "\n\n" +
			(_("If you already had Syncthing daemon configured, please, "
			   "exit this wizard and check your %s folder") % config_folder_link )
		), 0, 0, 1, 1)

class FindDaemonPage(Page):
	# Next page, "Download Daemon" is displayed only if needed.
	# When that happens, it becomes page with longest title and wizard
	# window changes size to accommodate this change. And i don't like
	# that.
	# To prevent this 'window jumping', padding is added here, so
	# this page is always one with longest name.
	TITLE = "Find Daemon"
	TYPE = Gtk.AssistantPageType.PROGRESS
	def init_page(self):
		""" Displayed while Syncthing binary is being searched for """
		self.label = WrappedLabel(
			"<b>" + _("Searching for Syncthing daemon.") + "</b>" +
			"\n\n" +
			_("Please wait...")
		)
		self.paths = []
		self.version_string = "v0.0"
		self.ignored_version = None
		self.attach(self.label, 0, 0, 1, 1)
	
	def prepare(self):
		default_path, default_binary = os.path.split(
				self.parent.config.get_default_value("syncthing_binary"))
		self.paths = []
		if default_path:
			self.paths += [ default_path ]
		self.paths += [ "./" ]
		self.paths += [ os.path.expanduser("~/.local/bin") ]
		self.paths += [ self.parent.st_configdir ]
		if is_portable():
			self.paths += [ ".\\data" ]
		if StDownloader is None:
			self.binaries = ["syncthing"]
		else:
			suffix, trash = StDownloader.determine_platform()
			self.binaries = ["syncthing", "syncthing%s" % (suffix,)]
			if suffix == "x64":
				# Allow 32bit binary on 64bit
				self.binaries += ["syncthing.x86"]
			if default_binary not in self.binaries:
				self.binaries = [ default_binary ] + self.binaries
		if IS_WINDOWS:
			self.paths += [ "c:/Program Files/syncthing",
				"c:/Program Files (x86)/syncthing",
				self.parent.st_configdir
				]
			self.binaries = ("syncthing.exe",)
		if "PATH" in os.environ:
			self.paths += os.environ["PATH"].split(":")
		log.info("Searching for syncthing binary...")
		GLib.idle_add(self.search)
	
	def search(self):
		"""
		Called repeatedly through GLib.idle_add, until binary is found
		or all possible paths are tried.
		"""
		try:
			path, self.paths = self.paths[0], self.paths[1:]
		except IndexError:
			# Out of possible paths. Not found
			if IS_WINDOWS:
				# On Windows, don't say anything and download Syncthing
				# directly
				self.parent.insert_and_go(DownloadSTPage())
				return False
			elif StDownloader is None:
				# On Linux with updater disabled, generate and
				# display error page
				title = _("Syncthing daemon not found.")
				message = _("Please, use package manager to install the Syncthing package.")
				page = self.parent.error(self, title, message, False)
				page.show_all()
				return False
			else:
				# On Linux with updater generate similar display error
				# and offer download
				from syncthing_gtk.app import MIN_ST_VERSION
				target_folder_link = '<a href="file://%s">%s</a>' % (
						os.path.expanduser(StDownloader.get_target_folder()),
						StDownloader.get_target_folder())
				dll_link = '<a href="https://github.com/syncthing/syncthing/releases">' + \
						_('download latest binary') + '</a>'
				message, title = "", None
				if self.ignored_version == None:
					# No binary was found
					title = _("Syncthing daemon not found.")
					message += _("Please, use package manager to install the Syncthing package "
								 "or %(download_link)s from Syncthing page and save it to your "
								 "%(target)s directory.") % {
						'download_link' : dll_link,
						'target' : target_folder_link
					}
				else:
					# Binary was found, but it was too old to be usable
					title = _("Syncthing daemon is too old.")
					message += _("Syncthing-GTK needs Syncthing daemon %(min)s or newer, but only %(actual)s were found.") % {
						'min' : MIN_ST_VERSION,
						'actual' : self.ignored_version
					}
					message += "\n"
					message += _("Please, use package manager to install the Syncthing package "
								 "or %(download_link)s from Syncthing page and save it to your "
								 "%(target)s directory.") % {
						'download_link' : dll_link,
						'target' : target_folder_link
					}
				message += "\n\n"
				message += _("Alternatively, Syncthing-GTK can download Syncthing binary") + " "
				message += _("to %s and keep it up-to-date, but this option is meant as") % \
							(target_folder_link,) + " "
				message += _("last resort and generally not suggested.")
				page = self.parent.error(self, title, message, False)
				# Attach [ ] Download Syncthing checkbox
				cb = Gtk.CheckButton(_("_Download Syncthing binary"), use_underline=True)
				cb.connect("toggled", lambda cb, *a : self.parent.set_page_complete(page, cb.get_active()))
				page.attach(cb,	0, 2, 2, 1)
				# Attach [ ] Autoupdate checkbox
				cb = Gtk.CheckButton(_("Auto_update downloaded binary"), use_underline=True)
				cb.connect("toggled", lambda cb, *a : self.parent.config.set("st_autoupdate", cb.get_active()))
				page.attach(cb,	0, 3, 2, 1)
				page.show_all()
				# Add Download page
				self.parent.insert(DownloadSTPage())
				return False
		
		for bin in self.binaries:
			bin_path = os.path.join(path, bin)
			log.info(" ... %s", bin_path)
			if os.path.isfile(bin_path):
				if os.access(bin_path, os.X_OK):
					# File exists and is executable, run it and parse
					# version string from output
					log.info("Binary found in %s", bin_path)
					if IS_WINDOWS: bin_path = bin_path.replace("/", "\\")
					p = DaemonProcess([ bin_path, '-version' ])
					p.connect('line', self.cb_process_output)
					p.connect('exit', self.cb_process_exit)
					p.connect('failed', self.cb_process_exit)
					p.start()
					return False
				else:
					log.info("Binary in %s is not not executable", bin_path)
		return True
	
	def cb_process_output(self, process, line):
		"""
		Called when daemon binary outputs line while it's being asked
		for version string.
		"""
		try:
			# Parse version string
			v = line.split(" ")[1]
			if v.startswith("v"):
				self.version_string = v
		except Exception:
			# Not line with version string, probably some other output
			pass
	
	def cb_process_exit(self, process, *a):
		""" Called after daemon binary outputs version and exits """
		from syncthing_gtk.app import MIN_ST_VERSION
		bin_path = process.get_commandline()[0]
		if compare_version(self.version_string, MIN_ST_VERSION):
			# Daemon binary exists, is executable and meets
			# version requirements. That's good, btw.
			self.parent.config["syncthing_binary"] = bin_path
			if not can_upgrade_binary(bin_path):
				# Don't try enable auto-update if binary is in
				# non-writable location (auto-update is enabled
				# by default on Windows only)
				self.parent.config["st_autoupdate"] = False
			self.parent.set_page_complete(self, True)
			self.label.set_markup(
					"<b>" + _("Syncthing daemon binary found.") + "</b>" +
					"\n\n" +
					_("Binary path:") + " " + bin_path + "\n" +
					_("Version:") + " " + self.version_string
				)
		else:
			# Found daemon binary too old to be usable.
			# Just ignore it and try to find better one.
			log.info("Binary in %s is too old", bin_path)
			self.ignored_version = self.version_string
			GLib.idle_add(self.search)

class DownloadSTPage(Page):
	TYPE = Gtk.AssistantPageType.PROGRESS
	TITLE = "Download Daemon"
	
	def init_page(self):
		""" Displayed while wizard downloads and extracts daemon """
		self.label = WrappedLabel("<b>" + _("Downloading Syncthing daemon.") + "</b>")
		self.version = WrappedLabel(_("Please wait..."))
		self.pb = Gtk.ProgressBar()
		self.label.props.margin_bottom = 15
		self.target = None
		self.attach(self.label,		0, 0, 1, 1)
		self.attach(self.version,	0, 1, 1, 1)
		self.attach(self.pb,		0, 2, 1, 1)
	
	def prepare(self):
		# Determine which Syncthing to use
		suffix, tag = StDownloader.determine_platform()
		# Report error on unsupported platforms
		if suffix is None or tag is None:
			pd = "%s %s %s" % (
				platform.uname()[0], platform.uname()[2],	# OS, version
				platform.uname()[4])						# architecture
			self.parent.error(self,
				_("Cannot download Syncthing daemon."),
				_("This platform (%s) is not supported") % (pd,),
				False)
			return
		# Determine target file & directory
		self.target = os.path.join(
			os.path.expanduser(StDownloader.get_target_folder()),
			"syncthing%s" % (suffix,)
			)
		# Create downloader and connect events
		self.sd = StDownloader(self.target, tag)
		self.sd.connect("error", self.on_download_error)
		self.sd.connect("version", self.on_version)
		self.sd.connect("download-progress", self.on_progress)
		self.sd.connect("download-finished", self.on_extract_start)
		self.sd.connect("extraction-progress", self.on_progress)
		self.sd.connect("extraction-finished", self.on_extract_finished)
		# Start downloading
		self.sd.get_version()
	
	def on_download_error(self, downloader, error, message):
		"""
		Called when download fails. This is fatal for now, user can
		only observe message, cry and quit program.
		"""
		message = "%s\n%s" % (
			str(error) if not error is None else "",
			message if not message is None else ""
			)
		self.parent.error(self,
			_("Failed to download Syncthing daemon package."),
			message, False)
		return
	
	def on_version(self, dowloader, version):
		self.version.set_markup("Downloading %s..." % (version, ))
		dowloader.download()
	
	def on_extract_start(self, *a):
		self.version.set_markup("Extracting...")
	
	def on_progress(self, dowloader, progress):
		self.pb.set_fraction(progress)
	
	def on_extract_finished(self, *a):
		""" Called after extraction is finished """
		# Everything done. Praise supernatural entities...
		self.label.set_markup("<b>" + _("Download finished.") + "</b>")
		self.parent.config["syncthing_binary"] = self.target
		self.version.set_markup(_("Binary path:") +
				" " + self.target)
		self.pb.set_visible(False)
		self.parent.set_page_complete(self, True)
	
class GenerateKeysPage(Page):
	TYPE = Gtk.AssistantPageType.PROGRESS
	TITLE = "Generate Keys"
	def init_page(self):
		""" Displayed while Syncthing binary is being searched for """
		self.label = WrappedLabel(
			"<b>%s</b>\n\n%s" % (
				_("Syncthing is generating RSA key and certificate."),
				_("This may take a while...")
			)
		)
		self.attach(self.label, 0, 0, 1, 1)
	
	def prepare(self):
		GLib.idle_add(self.start_binary)
	
	def start_binary(self):
		"""
		Starts Syncthing binary with -generate parameter and waits until
		key generation is finished
		"""
		self.parent.output_line("syncthing-gtk: Configuration directory: '%s'" % (self.parent.st_configdir,))
		# Create it, if needed
		try:
			os.makedirs(self.parent.st_configdir)
		except Exception as e:
			self.parent.output_line("syncthing-gtk: Failed to create configuration directory")
			self.parent.output_line("syncthing-gtk: %s" % (str(e),))
		# Run syncthing -generate
		self.parent.output_line("syncthing-gtk: Syncthing configuration directory: %s" % (self.parent.st_configdir,))
		self.process = DaemonProcess([ self.parent.config["syncthing_binary"], '-generate=%s' % self.parent.st_configdir ])
		self.process.connect('line', lambda proc, line : self.parent.output_line(line))
		self.process.connect('exit', self.cb_daemon_exit)
		self.process.connect('failed', self.cb_daemon_start_failed)
		self.process.start()
		return False
	
	def cb_daemon_start_failed(self, dproc, exception):
		self.parent.output_line("syncthing-gtk: Daemon startup failed")
		self.parent.output_line("syncthing-gtk: %s" % (str(exception),))
		self.cb_daemon_exit(dproc, -1)
	
	def cb_daemon_exit(self, dproc, exit_code):
		""" Called when Syncthing finishes """
		if exit_code == 0:
			# Finished without problem, advance to next page
			self.parent.set_page_complete(self, True)
			self.parent.next_page()
		else:
			self.parent.error(self,
				_("Failed to generate keys"),
				_("Syncthing daemon failed to generate RSA key or certificate."),
				True)

class HttpSettingsPage(Page):
	TYPE = Gtk.AssistantPageType.CONTENT
	TITLE = "Setup WebUI"
	def init_page(self):
		""" Permits user to set WebUI settings """
		# Wall of text
		label = WrappedLabel(
			"<b>" + _("WebUI setup") + "</b>" +
			"\n\n" +
			_("Syncthing can be managed remotely using WebUI and "
			  "even if you are going to use Syncthing-GTK, WebUI needs "
			  "to be enabled, as Syncthing-GTK uses it to communicate "
			  "with the Syncthing daemon.") +
			"\n\n" +
			_("If you prefer to be able to manage Syncthing remotely, "
			  "over the internet or on your local network, select <b>listen "
			  "on all interfaces</b> and set username and password to "
			  "protect Syncthing from unauthorized access.") +
			"\n" +
			_("Otherwise, select <b>listen on localhost</b>, so only "
			  "users and programs on this computer will be able to "
			  "interact with Syncthing.") +
			"\n"
		)
		# Radiobuttons
		lbl_radios = WrappedLabel("<b>" + _("WebUI Listen Addresses") + "</b>")
		self.rb_localhost = Gtk.RadioButton(label=_("Listen on _localhost"))
		self.rb_all_intfs = Gtk.RadioButton.new_from_widget(self.rb_localhost)
		self.rb_all_intfs.set_label(_("Listen on _all interfaces"))
		for x in (self.rb_localhost, self.rb_all_intfs):
			x.set_use_underline(True)
			x.set_property('margin-left', 15)
		# Username & password input boxes
		self.tx_username = Gtk.Entry()
		self.tx_password = Gtk.Entry()
		self.lbl_username = WrappedLabel(_("_Username"))
		self.lbl_password = WrappedLabel(_("_Password"))
		self.lbl_username.set_mnemonic_widget(self.tx_username)
		self.lbl_password.set_mnemonic_widget(self.tx_password)
		self.tx_password.set_visibility(False)
		self.tx_password.props.caps_lock_warning = True
		for x in (self.lbl_username, self.lbl_password):
			x.set_use_underline(True)
			x.set_property('margin-left', 45)
			x.set_property('margin-bottom', 5)
		for x in (self.tx_username, self.tx_password):
			x.set_property('margin-bottom', 5)
		# Connect signals
		for x in (self.rb_localhost, self.rb_all_intfs):
			x.connect("toggled", self.cb_stuff_changed)
		for x in (self.tx_username, self.tx_password):
			x.connect("changed", self.cb_stuff_changed)
			x.connect("delete-text", self.cb_stuff_changed)
			x.connect("insert-text", self.cb_stuff_changed)
		# Attach everything
		self.attach(label, 0, 0, 3, 1)
		self.attach(lbl_radios, 0, 1, 3, 1)
		self.attach(self.rb_localhost, 0, 2, 2, 1)
		self.attach(self.rb_all_intfs, 0, 3, 2, 1)
		self.attach(self.lbl_username, 0, 4, 1, 1)
		self.attach(self.lbl_password, 0, 5, 1, 1)
		self.attach(self.tx_username, 1, 4, 2, 1)
		self.attach(self.tx_password, 1, 5, 2, 1)
	
	def cb_stuff_changed(self, *a):
		""" Called every time user changes anything on this page """
		# Enable / disable username & password input boxes
		for x in (self.tx_username, self.tx_password, self.lbl_username, self.lbl_password):
			x.set_sensitive(self.rb_all_intfs.get_active())
		# Set page as 'complete' and store settings
		# if either localhost is selected, or username & password is filled
		values_ok = self.rb_localhost.get_active() or (len(self.tx_username.get_text().strip()) and len(self.tx_password.get_text().strip()))
		self.parent.set_page_complete(self, values_ok)
		if values_ok:
			if self.rb_localhost.get_active():
				self.parent.syncthing_options["listen_ip"] = "127.0.0.1"
			else:
				self.parent.syncthing_options["listen_ip"] = "0.0.0.0"
			self.parent.syncthing_options["user"] = str(self.tx_username.get_text())
			self.parent.syncthing_options["password"] = str(self.tx_password.get_text())
	
	def prepare(self):
		# Refresh UI
		self.cb_stuff_changed()

class SaveSettingsPage(Page):
	TYPE = Gtk.AssistantPageType.PROGRESS
	TITLE = "Save Settings"
	def init_page(self):
		""" Displayed while settings are being saved """
		self.label = WrappedLabel("<b>" + _("Saving settings...") + "</b>" + "\n\n")
		self.status = Gtk.Label(_("Checking for available port..."))
		self.attach(self.label,		0, 0, 1, 1)
		self.attach(self.status,	0, 1, 1, 1)
	
	def prepare(self):
		GLib.idle_add(self.check_port, DEFAULT_PORT)
	
	def check_port(self, port):
		"""
		Tries to open TCP port to check it availability.
		It this fails, checks next ports, until MAX_PORT is reached.
		When MAX_PORT is reached, it's safe to assume that something
		completely wrong is happening and an error should be displayed.
		"""
		if port >= MAX_PORT:
			# Remove config.xml that I just created
			try:
				os.unlink(self.parent.st_configfile)
			except Exception as e:
				self.parent.output_line("syncthing-gtk: %s" % (str(e),))
			self.parent.error(self,
				_("Failed to find unused port for listening."),
				_("Please, check your firewall settings and try again."),
				False)
			return
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		try:
			s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
			s.bind((self.parent.syncthing_options["listen_ip"], port))
			s.listen(0)
			s.close()
			# Good, port is available
			self.parent.output_line("syncthing-gtk: chosen port %s" % (port,))
			self.port = port
			self.parent.syncthing_options["port"] = str(port)
			self.status.set_markup(_("Saving..."))
			GLib.idle_add(self.save_settings)
		except socket.error:
			# Address already in use (or some crazy error)
			self.parent.output_line("syncthing-gtk: port %s is not available" % (port,))
			GLib.idle_add(self.check_port, port + 1)
		finally:
			del s
	
	def ct_textnode(self, xml, parent, name, value):
		""" Helper method """
		el = xml.createElement(name)
		text = xml.createTextNode(value)
		el.appendChild(text)
		parent.appendChild(el)
	
	def save_settings(self):
		"""
		Loads&parses XML, changes some values and writes it back.
		No backup is created as this wizard is expected to be ran
		only if there is no config in first place.
		"""
		# Generate API key
		self.apikey = "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(30))
		log.debug("Generated apikey %s", self.apikey)
		xml = None
		try:
			# Load XML file
			with open(self.parent.st_configfile, "r") as conf_file:
				config = conf_file.read()
			xml = minidom.parseString(config)
		except Exception as e:
			self.parent.output_line("syncthing-gtk: %s" % (traceback.format_exc(),))
			self.parent.error(self,
				_("Failed to load Syncthing configuration"),
				str(e),
				True)
			return False
		try:
			# Prepare elements
			gui = xml.getElementsByTagName("configuration")[0] \
					.getElementsByTagName("gui")[0]
			au = xml.getElementsByTagName("configuration")[0] \
					.getElementsByTagName("options")[0] \
					.getElementsByTagName("autoUpgradeIntervalH")[0]
			while gui.firstChild != None:
				gui.removeChild(gui.firstChild)
			# Update data
			self.ct_textnode(xml, gui, "address", "%s:%s" % (
							self.parent.syncthing_options["listen_ip"],
							self.parent.syncthing_options["port"],
					))
			self.ct_textnode(xml, gui, "user", self.parent.syncthing_options["user"])
			self.ct_textnode(xml, gui, "password", bcrypt.hashpw(
				self.parent.syncthing_options["password"].encode('utf8'),
				bcrypt.gensalt()
			).decode('utf8'))
			self.ct_textnode(xml, gui, "apikey", self.apikey)
			gui.setAttribute("enabled", "true")
			gui.setAttribute("tls", "false")
			au.firstChild.replaceWholeText("0")
		except Exception as e:
			self.parent.output_line("syncthing-gtk: %s" % (traceback.format_exc(),))
			self.parent.error(self,
				_("Failed to modify Syncthing configuration"),
				str(e),
				True)
			return False
		try:
			# Write XML back to file
			with open(self.parent.st_configfile, "w", encoding="utf-8") as f:
				f.write(xml.toxml())
		except Exception as e:
			self.parent.output_line("syncthing-gtk: %s" % (traceback.format_exc(),))
			self.parent.error(self,
				_("Failed to save Syncthing configuration"),
				str(e),
				True)
			return False
		self.parent.set_page_complete(self, True)
		self.parent.next_page()
		return False

class LastPage(GenerateKeysPage):
	TYPE = Gtk.AssistantPageType.SUMMARY
	TITLE = "Finish"
	def init_page(self):
		""" Well, it's last page. """
		label = WrappedLabel(
			"<b>" + _("Done.") + "</b>" +
			"\n\n" +
			_("Syncthing has been successfully configured.") +
			"\n" +
			_("You can configure more details later, in "
			  "<b>UI Settings</b> and <b>Daemon Settings</b> menus "
			  "in main window of application.")
		)
		self.attach(label, 0, 0, 1, 1)
	
	def prepare(self):
		# Configure main app to manage Syncthing daemon by default
		self.parent.config["autostart_daemon"] = 1
		self.parent.config["autokill_daemon"] = 1
		self.parent.config["minimize_on_start"] = False
		if IS_WINDOWS:
			self.parent.config["use_old_header"] = True
		self.parent.quit_button.get_parent().remove(self.parent.quit_button)
		self.parent.finished = True



#!/usr/bin/env python3
"""
Syncthing-GTK - 1st run wizard

Basically runs syncthing daemon with -generate option and setups some
values afterwards.
"""


from syncthing_gtk.editordialog import EditorDialog
from syncthing_gtk.stdownloader import StDownloader
from syncthing_gtk.tools import IS_WINDOWS, IS_XP
from syncthing_gtk.tools import _ # gettext function
from syncthing_gtk.uisettingsdialog import UISettingsDialog, browse_for_binary
import os, platform

VALUES = [ "vsyncthing_binary" ]

class FindDaemonDialog(EditorDialog):
	RESPONSE_SAVED = 1
	RESPONSE_QUIT = 2
	def __init__(self, app):
		EditorDialog.__init__(self, app, "find-daemon.glade",
			_("Can't invoke the daemon"))
		self.app = app
		exe = "syncthing.exe" if IS_WINDOWS else _("Syncthing binary")
		self["lblMessage"].set_markup("%s\n%s" % (
			_("Syncthing daemon binary cannot be found."),
			_("If you have Syncthing installed, please, set path to "
			  "%s below or click on <b>Download</b> "
			  "button to download latest Syncthing package.") % (exe,)
		))
		if IS_XP or StDownloader is None:
			# Downloading is not offered on XP (github will not talk to it)
			# or if StDownloader module is not packaged
			self["lblMessage"].set_markup("%s\n%s" % (
				_("Syncthing daemon binary cannot be found."),
				_("If you have Syncthing installed, please, set path to "
				  "%s below") % (exe,)
			))
			self.hide_download_button()
	
	### Dialog emulation
	def set_transient_for(self, parent):
		self["editor"].set_transient_for(parent)
	
	def set_message(self, m):
		self["lblMessage"].set_markup(m)
	
	def hide_download_button(self):
		self["btDownload"].set_visible(False)
	
	def run(self):
		return self["editor"].run()
	
	def destroy(self):
		self.close()
	
	
	
	
	### UI callbacks
	def cb_btBrowse_clicked(self, *a):
		""" Display file browser dialog to browse for syncthing binary """
		browse_for_binary(self["editor"], self, "vsyncthing_binary")
	
	def cb_btDownload_clicked(self, *a):
		"""
		Disable half of dialog and start downloading syncthing package
		"""
		# Determine which syncthing to use
		suffix, tag = StDownloader.determine_platform()
		# Report error on unsupported platforms
		if suffix is None or tag is None:
			# Disable download button
			self["btDownload"].set_sensitive(False)
			# Set message
			pd = "%s %s" % (
				platform.uname()[0],	# OS
				platform.uname()[4])	# architecture
			self["lblDownloadProgress"].set_markup("%s %s" % (
					_("Cannot download Syncthing daemon."),
					_("This platform (%s) is not supported") % (pd,),
				))
			return
		# Determine target file & directory
		self.target = os.path.join(
			os.path.expanduser(StDownloader.get_target_folder()),
			"syncthing%s" % (suffix,)
			)
		# Create downloader and connect events
		sd = StDownloader(self.target, tag)
		sd.connect("error", self.cb_download_error)
		sd.connect("version", self.cb_version)
		sd.connect("download-progress", self.cb_progress)
		sd.connect("download-finished", self.cb_extract_start)
		sd.connect("extraction-progress", self.cb_progress)
		sd.connect("extraction-finished", self.cb_extract_finished)
		# Display message and start downloading
		self["lblDownloadProgress"].set_markup(_("Downloading..."))
		self["btDownload"].set_visible(False)
		self["pbDownload"].set_visible(True)
		self["vsyncthing_binary"].set_sensitive(False)
		self["btBrowse"].set_sensitive(False)
		self["btSave"].set_sensitive(False)
		sd.get_version()
	
	def cb_btQuit_clicked(self, *a):
		""" Handler for 'Quit' button """
		self["editor"].response(FindDaemonDialog.RESPONSE_QUIT)
	
	def cb_bt_ui_settings_clicked(self, *a):
		""" Handler for 'UI Settings' button """
		e = UISettingsDialog(self.app)
		e.connect('close', self.cb_ui_settings_closed)
		e.load()
		e.show(self["window"])
	
	def cb_ui_settings_closed(self, *a):
		self.load()
	
	### EditorDialog overrides
	
	#@Overrides
	def load_data(self):
		# Don't load data from syncthing daemon, it knows nothing...
		copy = { k : self.app.config[k] for k in self.app.config }
		self.cb_data_loaded(copy)
		self.cb_check_value()
	
	#@Overrides
	def on_data_loaded(self):
		self.values = self.config
		self.checks = {}
		return self.display_values(VALUES)
	
	#@Overrides
	def update_special_widgets(self, *a):
		pass
	
	#@Overrides
	def on_save_requested(self):
		self.store_values(VALUES)
		# Save data to configuration file
		for k in self.values:
			self.app.config[k] = self.values[k]
		# Report work done
		self.syncthing_cb_post_config()
	
	#@Overrides
	def on_saved(self):
		self["editor"].response(FindDaemonDialog.RESPONSE_SAVED)
	
	
	
	
	### Downloader callbacks
	def cb_download_error(self, downloader, error, message):
		"""
		Called when download fails. User can click 'Download' to
		try it again.
		"""
		self["lblDownloadProgress"].set_markup(_("Download failed."))
		self["btDownload"].set_visible(True)
		self["pbDownload"].set_visible(False)
		self["vsyncthing_binary"].set_sensitive(True)
		self["btBrowse"].set_sensitive(True)
		self["btSave"].set_sensitive(True)
	
	def cb_version(self, downloader, version):
		self["lblDownloadProgress"].set_markup("Downloading %s..." % (version, ))
		downloader.download()
	
	def cb_extract_start(self, *a):
		self["lblDownloadProgress"].set_markup("Extracting...")
	
	def cb_progress(self, downloader, progress):
		self["pbDownload"].set_fraction(progress)
	
	def cb_extract_finished(self, downloader, *a):
		""" Called after extraction is finished """
		self["vsyncthing_binary"].set_sensitive(True)
		self["btBrowse"].set_sensitive(True)
		self["vsyncthing_binary"].set_text(downloader.get_target())
		self["lblDownloadProgress"].set_markup("<b>" + _("Download finished.") + "</b>")
		self["pbDownload"].set_visible(False)
		self["btSave"].set_sensitive(True)
	



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Syncthing-GTK - StatusIcon

"""


import locale
import os
import sys
import logging

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk

from syncthing_gtk.tools import IS_UNITY, IS_KDE, IS_CINNAMON, IS_LXQT
from syncthing_gtk.tools import _ # gettext function

log = logging.getLogger("StatusIcon")


#                | KDE5            | MATE      | Unity      | Cinnamon   | Cairo-Dock (classic) | Cairo-Dock (modern) | KDE4      |
#----------------+-----------------+-----------+------------+------------+----------------------+---------------------+-----------+
# StatusIconQt5  | very good (KF5) | -         | -          | -          | -                    | -                   | -         |
# StatusIconAppI | good²           | none      | excellent  | none       | none                 | excellent           | good²     |
# StatusIconGTK3 | good            | excellent | none       | very good¹ | very good¹           | none                | good⁴     |
#
# Notes:
#  - StatusIconQt5:
#     - It's pretty unstable and leads to crashes
#     - Only tested on Qt 5.4 which only supports Qt5 through a KDE frameworks plugin
#  - StatusIconAppIndicator does not implement any fallback (but the original libappindicator did)
#  - Markers:
#     ¹ Icon cropped
#     ² Does not support left-click
#     ³ It works, but looks ugly and does not support left-click
#     ⁴ Does not support icon states
#     ⁵ For some menu items the standard GTK icons are used instead of the monotone ones


class StatusIcon(GObject.GObject):
	"""
	Base class for all status icon backends
	"""
	TRAY_TITLE     = _("Syncthing")
	
	__gsignals__ = {
		"clicked": (GObject.SIGNAL_RUN_FIRST, None, ()),
	}
	
	__gproperties__ = {
		"active": (
			GObject.TYPE_BOOLEAN,
			"is the icon user-visible?",
			"does the icon back-end think that anything is might be shown to the user?",
			True,
			GObject.PARAM_READWRITE
		)		
	}
	
	def __init__(self, icon_path, popupmenu, force=False):
		GObject.GObject.__init__(self)
		self.__icon_path = os.path.normpath(os.path.abspath(icon_path))
		self.__popupmenu = popupmenu
		self.__active    = True
		self.__visible   = False
		self.__hidden    = False
		self.__icon      = "si-syncthing-unknown"
		self.__text      = ""
		self.__force     = force
	
	def get_active(self):
		"""
		Return whether there is at least a chance that the icon might be shown to the user
		
		If this returns `False` then the icon will definetely not be shown, but if it returns `True` it doesn't have to
		be visible...
		
		<em>Note:</em> This value is not directly influenced by calling `hide()` and `show()`.
		
		@return {bool}
		"""
		return self.get_property("active")
	
	def set(self, icon=None, text=None):
		"""
		Set the status icon image and descriptive text
		
		If either of these are `None` their previous value will be used.
		
		@param {String} icon
		       The name of the icon to show (i.e. `si-syncthing-idle`)
		@param {String} text
		       Some text that indicates what the application is currently doing (generally this be used for the tooltip)
		"""
		if IS_KDE and isinstance(self, StatusIconDBus) and not icon.startswith("si-syncthing"):
			# KDE seems to be the only platform that has proper support for icon states
			# (all other implementations just hide the icon completely when its passive)
			self.__visible = False
		elif not icon.endswith("-0"): # si-syncthing-0
			# Ignore first syncing icon state to prevent the icon from flickering
			# into the main notification bar during initialization
			self.__visible = True
		
		if self.__hidden:
			self._set_visible(False)
		else:
			self._set_visible(self.__visible)
	
	def hide(self):
		"""
		Hide the icon
		
		This method tries its best to ensure the icon is hidden, but there are no guarantees as to how use well its
		going to work.
		"""
		self.__hidden = True
		self._set_visible(False)
	
	def show(self):
		"""
		Show a previously hidden icon
		
		This method tries its best to ensure the icon is hidden, but there are no guarantees as to how use well its
		going to work.
		"""
		self.__hidden = False
		self._set_visible(self.__visible)
	
	
	def _is_forced(self):
		return self.__force
	
	def _on_click(self, *a):
		self.emit("clicked")
	
	def _get_icon(self, icon=None):
		"""
		@internal
		
		Use `set()` instead.
		"""
		if icon:
			self.__icon = icon
		return self.__icon
	
	def _get_text(self, text=None):
		"""
		@internal
		
		Use `set()` instead.
		"""
		if text:
			self.__text = text
		return self.__text
	
	def _get_popupmenu(self):
		"""
		@internal
		"""
		return self.__popupmenu
	
	def _set_visible(self, visible):
		"""
		@internal
		"""
		pass

	def do_get_property(self, property):
		if property.name == "active":
			return self.__active
		else:
			raise AttributeError("Unknown property %s" % property.name)
	
	def do_set_property(self, property, value):
		if property.name == "active":
			self.__active = value
		else:
			raise AttributeError("unknown property %s" % property.name)


class StatusIconDummy(StatusIcon):
	"""
	Dummy status icon implementation that does nothing
	"""
	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)
		
		# Pretty unlikely that this will be visible...
		self.set_property("active", False)
		if IS_UNITY or IS_KDE:
			log.warning("Failed to load modules required for status icon. "
						"Please, make sure libappindicator package and python "
						"bindings are installed.")
		else:
			log.warning("Failed to load modules required for status icon")
	
	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)
		
		self._get_icon(icon)
		self._get_text(text)


class StatusIconGTK3(StatusIcon):
	"""
	Gtk.StatusIcon based status icon backend
	"""
	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)
		
		if not self._is_forced():
			if IS_UNITY:
				# Unity fakes SysTray support but actually hides all icons...
				raise NotImplementedError
		
		self._tray = Gtk.StatusIcon()
		
		self._tray.connect("activate", self._on_click)
		self._tray.connect("popup-menu", self._on_rclick)
		self._tray.connect("notify::embedded", self._on_embedded_change)
		
		self._tray.set_visible(True)
		self._tray.set_name("syncthing-gtk")
		self._tray.set_title(self.TRAY_TITLE)
		
		# self._tray.is_embedded() must be called asynchronously
		# See: http://stackoverflow.com/a/6365904/277882
		GLib.idle_add(self._on_embedded_change)
	
	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)
		
		self._tray.set_from_icon_name(self._get_icon(icon))
		self._tray.set_tooltip_text(self._get_text(text))
	
	def _on_embedded_change(self, *args):
		# Without an icon update at this point GTK might consider the icon embedded and visible even through
		# it can't actually be seen...
		self._tray.set_from_icon_name(self._get_icon())
		
		# An invisible tray icon will never be embedded but it also should not be replaced
		# by a fallback icon
		is_embedded = self._tray.is_embedded() or not self._tray.get_visible()
		# On some desktops, above check fails but tray is always visible
		is_embedded = is_embedded or IS_KDE or IS_LXQT or IS_CINNAMON
		if is_embedded != self.get_property("active"):
			self.set_property("active", is_embedded)
	
	def _on_rclick(self, si, button, time):
		self._get_popupmenu().popup(None, None, None, None, button, time)
	
	def _set_visible(self, active):
		StatusIcon._set_visible(self, active)
		
		self._tray.set_visible(active)


class StatusIconDBus(StatusIcon):
	pass


class StatusIconQt(StatusIconDBus):
	"""
	Base implementation for all Qt-based backends that provides GMenu to QMenu conversion services
	"""
	def _make_qt_action(self, menu_child_gtk, menu_qt):
		# This is a separate function to make sure that the Qt callback function are executed
		# in the correct `locale()` context and do net trigger events on the wrong Gtk menu item
		
		# Create menu item
		action = self._qt_types["QAction"](menu_qt)
		
		# Convert item to separator if appropriate
		action.setSeparator(isinstance(menu_child_gtk, Gtk.SeparatorMenuItem))
		
		# Copy sensitivity
		def set_sensitive(*args):
			action.setEnabled(menu_child_gtk.is_sensitive())
		menu_child_gtk.connect("notify::sensitive", set_sensitive)
		set_sensitive()
		
		# Copy checkbox state
		if isinstance(menu_child_gtk, Gtk.CheckMenuItem):
			action.setCheckable(True)
			def _set_visible(*args):
				action.setChecked(menu_child_gtk.get_active())
			menu_child_gtk.connect("notify::active", _set_visible)
			_set_visible()
		
		# Copy icon
		if isinstance(menu_child_gtk, Gtk.ImageMenuItem):
			def set_image(*args):
				image = menu_child_gtk.get_image()
				if image and image.get_storage_type() == Gtk.ImageType.PIXBUF:
					# Converting GdkPixbufs to QIcons might be a bit inefficient this way,
					# but it requires only very little code and looks very stable
					png_buffer = image.get_pixbuf().save_to_bufferv("png", [], [])[1]
					image      = self._qt_types["QImage"].fromData(png_buffer)
					pixmap     = self._qt_types["QPixmap"].fromImage(image)
					
					action.setIcon(self._qt_types["QIcon"](pixmap))
				elif image:
					icon_name = None
					if image.get_storage_type() == Gtk.ImageType.ICON_NAME:
						icon_name = image.get_icon_name()[0]
					if image.get_storage_type() == Gtk.ImageType.STOCK:
						icon_name = image.get_stock()[0]
					
					action.setIcon(self._get_icon_by_name(icon_name))
				else:
					action.setIcon(self._get_icon_by_name(None))
			menu_child_gtk.connect("notify::image", set_image)
			set_image()
		
		# Set label
		def set_label(*args):
			label = menu_child_gtk.get_label()
			if isinstance(menu_child_gtk, Gtk.ImageMenuItem) and menu_child_gtk.get_use_stock():
				label = Gtk.stock_lookup(label).label
			if isinstance(label, str):
				label = label.decode(locale.getpreferredencoding())
			if menu_child_gtk.get_use_underline():
				label = label.replace("_", "&")
			action.setText(label)
		menu_child_gtk.connect("notify::label", set_label)
		set_label()
		
		# Add submenus
		def set_popupmenu(*args):
			action.setMenu(self._get_popupmenu(menu_child_gtk.get_submenu()))
		menu_child_gtk.connect("notify::popupmenu", set_popupmenu)
		set_popupmenu()
		
		# Hook up Qt signals to their GTK counterparts
		action.triggered.connect(lambda *a: menu_child_gtk.emit("activate"))
		
		return action
	
	def _get_icon_by_name(self, icon_name):
		if icon_name:
			icon_file = self._gtk_icon_theme.lookup_icon(icon_name, 48, 0)
			if not icon_file:
				log.info("Skipping unknown icon file: %s" % (icon_name))
				return self._qt_types["QIcon"]()
			
			icon_path = icon_file.get_filename()
			if not icon_path:
				return self._qt_types["QIcon"]()
			
			icon_dir, icon_basename = os.path.split(os.path.realpath(icon_path))
		
			# If we don't resolve all icon names (i.e.: realpath) before passing them to Qt
			# SOME OF THEM will be dropped (especially if their name started with "gtk-" originally)
			icon_name = os.path.splitext(icon_basename)[0]
		
			# Make sure that Qt can find this icon by its name, by adding
			# the directory to the icon theme search path
			# This extra step is required because we have to set the application
			# style to "motif" during Qt initialization
			if icon_dir not in self._qt_types["QIcon"].themeSearchPaths():
				theme_search_paths = self._qt_types["QIcon"].themeSearchPaths()
				theme_search_paths.prepend(icon_dir)
				self._qt_types["QIcon"].setThemeSearchPaths(theme_search_paths)
		
			return self._qt_types["QIcon"].fromTheme(icon_name, self._qt_types["QIcon"](icon_path))
		
		return self._qt_types["QIcon"]()
	
	def _set_qt_types(self, **kwargs):
		self._gtk_icon_theme = Gtk.IconTheme.get_default()
		
		self._qt_types = kwargs
	
	def _get_popupmenu(self, menu_gtk=False):
		menu_gtk = menu_gtk if menu_gtk is not False else StatusIcon._get_popupmenu(self)
		if not menu_gtk:
			return None
		
		menu_qt = self._qt_types["QMenu"]()
		for menu_child_gtk in menu_gtk.get_children():
			menu_qt.addAction(self._make_qt_action(menu_child_gtk, menu_qt))
		
		return menu_qt

class StatusIconAppIndicator(StatusIconDBus):
	"""
	Unity's AppIndicator3.Indicator based status icon backend
	"""
	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)
		
		try:
			import gi
			gi.require_version('AyatanaAppIndicator3', '0.1')
			from gi.repository import AyatanaAppIndicator3 as appindicator
		except (ImportError, ValueError):
			try:
				import gi
				gi.require_version('AppIndicator3', '0.1')
				from gi.repository import AppIndicator3 as appindicator
			except (ImportError, ValueError):
				raise NotImplementedError
		
		self._status_active  = appindicator.IndicatorStatus.ACTIVE
		self._status_passive = appindicator.IndicatorStatus.PASSIVE

		category = appindicator.IndicatorCategory.APPLICATION_STATUS
		# Whatever icon is set here will be used as a tooltip icon during the entire time to icon is shown
		self._tray = appindicator.Indicator.new("syncthing-gtk", self._get_icon(), category)
		self._tray.set_menu(self._get_popupmenu())
		self._tray.set_title(self.TRAY_TITLE)
	
	def _set_visible(self, active):
		StatusIcon._set_visible(self, active)
		
		self._tray.set_status(self._status_active if active else self._status_passive)
	
	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)
		
		self._tray.set_icon_full(self._get_icon(icon), self._get_text(text))



class StatusIconProxy(StatusIcon):
	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)
		
		self._arguments  = (args, kwargs)
		self._status_fb  = None
		self._status_gtk = None
		self.set("si-syncthing-unknown", "")
		
		# Do not ever force-show indicators when they do not think they'll work
		if "force" in self._arguments[1]:
			del self._arguments[1]["force"]
		
		try:
			# Try loading GTK native status icon
			self._status_gtk = StatusIconGTK3(*args, **kwargs)
			self._status_gtk.connect("clicked",        self._on_click)
			self._status_gtk.connect("notify::active", self._on_notify_active_gtk)
			self._on_notify_active_gtk()
			
			log.info("Using backend StatusIconGTK3 (primary)")
		except NotImplementedError:
			# Directly load fallback implementation
			self._load_fallback()
	
	def _on_click(self, *args):
		self.emit("clicked")
	
	def _on_notify_active_gtk(self, *args):
		if self._status_fb:
			# Hide fallback icon if GTK icon is active and vice-versa
			if self._status_gtk.get_active():
				self._status_fb.hide()
			else:
				self._status_fb.show()
		elif not self._status_gtk.get_active():
			# Load fallback implementation
			self._load_fallback()
	
	def _on_notify_active_fb(self, *args):
		active = False
		if self._status_gtk and self._status_gtk.get_active():
			active = True
		if self._status_fb and self._status_fb.get_active():
			active = True
		self.set_property("active", active)
	
	def _load_fallback(self):
		if IS_KDE:
			status_icon_backends = [StatusIconAppIndicator, StatusIconQt5, StatusIconDummy]
		else:
			status_icon_backends = [StatusIconAppIndicator, StatusIconDummy]
		
		if not self._status_fb:
			for StatusIconBackend in status_icon_backends:
				try:
					self._status_fb = StatusIconBackend(*self._arguments[0], **self._arguments[1])
					self._status_fb.connect("clicked",        self._on_click)
					self._status_fb.connect("notify::active", self._on_notify_active_fb)
					self._on_notify_active_fb()
					
					log.warning("StatusIcon: Using backend %s (fallback)" % StatusIconBackend.__name__)
					break
				except NotImplementedError:
					continue
		
			# At least the dummy backend should have been loaded at this point...
			assert self._status_fb
		
		# Update fallback icon
		self.set(self._icon, self._text)
	
	def set(self, icon=None, text=None):
		self._icon = icon
		self._text = text
		
		if self._status_gtk:
			self._status_gtk.set(icon, text)
		if self._status_fb:
			self._status_fb.set(icon, text)
	
	def hide(self):
		if self._status_gtk:
			self._status_gtk.hide()
		if self._status_fb:
			self._status_fb.hide()
	
	def show(self):
		if self._status_gtk:
			self._status_gtk.show()
		if self._status_fb:
			self._status_fb.show()

def get_status_icon(*args, **kwargs):
	# Try selecting backend based on environment variable
	if "SYNCTHING_STATUS_BACKEND" in os.environ:
		kwargs["force"] = True
		
		status_icon_backend_name = "StatusIcon%s" % (os.environ.get("SYNCTHING_STATUS_BACKEND"))
		if status_icon_backend_name in globals():
			try:
				status_icon = globals()[status_icon_backend_name](*args, **kwargs)
				log.info("StatusIcon: Using requested backend %s" % (status_icon_backend_name))
				return status_icon
			except NotImplementedError:
				log.error("StatusIcon: Requested backend %s is not supported" % (status_icon_backend_name))
		else:
			log.error("StatusIcon: Requested backend %s does not exist" % (status_icon_backend_name))
		
		return StatusIconDummy(*args, **kwargs)
	
	# Use proxy backend to determine the correct backend while the application is running
	return StatusIconProxy(*args, **kwargs)



#!/usr/bin/env python3
"""
Syncthing-GTK - Configuration

Configuration object implementation; Uses JSON.
Config file is by default in ~/.config/syncthing-gtk/config.json
or other ~/.config equivalent
"""


from datetime import datetime
from syncthing_gtk.tools import IS_WINDOWS, get_config_dir, is_portable
import dateutil.parser
import os, sys, json, logging
log = logging.getLogger("Configuration")

LONG_AGO = datetime.fromtimestamp(1)

class _Configuration(object):
	"""
	Configuration object implementation.
	Use like dict to save / access values
	"""
	
	# Dict with keys that are required in configuration file
	# and default values for those keys.
	# Format: key : (type, default)
	REQUIRED_KEYS = {
		"autostart_daemon"			: (int, 2),	# 0 - wait for daemon, 1 - autostart, 2 - ask
		"autokill_daemon"			: (int, 2),	# 0 - never kill, 1 - always kill, 2 - ask
		"daemon_priority"			: (int, 0), # uses nice values
		"max_cpus"					: (int, 0), # 0 for all cpus
		"syncthing_binary"			: (str, "/usr/bin/syncthing"),
		"syncthing_arguments"		: (str, ""),
		"minimize_on_start"			: (bool, False),
		"folder_as_path"			: (bool, True),
		"use_old_header"			: (bool, False),
		"animate_icon"				: (bool, True),
		"notification_for_update"	: (bool, True),
		"notification_for_folder"	: (bool, False),
		"notification_for_error"	: (bool, True),
		"st_autoupdate"				: (bool, False),
		"last_updatecheck"			: (datetime, LONG_AGO),
		"window_position"			: (tuple, None),
		"infobox_style"				: (str, 'font_weight="bold" font_size="large"'),
		"icon_theme"				: (str, 'syncthing'),
		"force_dark_theme"			: (bool, False),	# Windows-only
		"language"					: (str, ""),		# Windows-only
		"file_browser"				: (str, "explore"),	# Windows-only
	}
	
	# Overrides some default values on Windows
	WINDOWS_OVERRIDE = {
		"syncthing_binary"			: (str, "C:\\Program Files\\Syncthing\\syncthing.exe"),
		"autokill_daemon"			: (int, 1),
		"use_old_header"			: (bool, False),
		"st_autoupdate"				: (bool, True),
	}
	
	def __init__(self):
		try:
			self.load()
		except Exception as e:
			log.warning("Failed to load configuration; Creating new one.")
			log.warning("Reason: %s", (e,))
			self.create()
		
		# Convert objects serialized as string back to object
		self.convert_values()
		# Check if everything is in place, add default value
		# where value is missing
		if self.check_values():
			# check_values returns True if any default value is added
			log.info("Saving configuration...")
			self.save()
	
	def load(self):
		# Check & create directory
		if not os.path.exists(self.get_config_dir()):
			try:
				os.makedirs(self.get_config_dir())
			except Exception as e:
				log.error("Cannot create configuration directory")
				log.exception(e)
				sys.exit(1)
		# Load json
		with open(self.get_config_file(), "r") as conf:
			self.values = json.loads(conf.read())

	def get_default_value(self, key):
		if IS_WINDOWS:
			return self.WINDOWS_OVERRIDE.get(key, self.REQUIRED_KEYS[key])[-1]
		return self.REQUIRED_KEYS[key][-1]
	
	def get_config_dir(self):
		return os.path.join(get_config_dir(), "syncthing-gtk")
	
	def get_config_file(self):
		return os.path.join(self.get_config_dir(), "config.json")
	
	def create(self):
		""" Creates new, empty configuration """
		self.values = {}
		self.check_values()
		self.save()
	
	def check_values(self):
		"""
		Check if all required values are in place and fill by default
		whatever is missing.
		
		Returns True if anything gets changed.
		"""
		needs_to_save = False
		for key in Configuration.REQUIRED_KEYS:
			tp, default = Configuration.REQUIRED_KEYS[key]
			if not self.check_type(key, tp):
				log.verbose("Configuration key %s is missing. Using default", key)
				if IS_WINDOWS and key in Configuration.WINDOWS_OVERRIDE:
					tp, default = Configuration.WINDOWS_OVERRIDE[key]
				self.values[key] = default
				needs_to_save = True
		return needs_to_save
	
	def convert_values(self):
		"""
		Converts all objects serialized as string back to object
		"""
		for key in Configuration.REQUIRED_KEYS:
			if key in self.values:
				tp, trash = Configuration.REQUIRED_KEYS[key]
				try:
					if tp == datetime and type(self.values[key]) == str:
						# Parse datetime
						self.values[key] = dateutil.parser.parse(self.values[key])
					elif tp == tuple and type(self.values[key]) == list:
						# Convert list to tuple
						self.values[key] = tuple(self.values[key])
					elif tp == bool and type(self.values[key]) == int:
						# Convert bools
						self.values[key] = bool(self.values[key])
				except Exception as e:
					log.warning("Failed to parse configuration value '%s'. Using default.", key)
					log.warning(e)
					# Value will be re-created by check_values method
					del self.values[key]
	
	def check_type(self, key, tp):
		"""
		Returns True if value is set and type match.
		Auto-converts objects serialized as string back to objects
		"""
		if not key in self.values:
			return False
		# Handle special cases
		if type(self.values[key]) == str and tp == str:
			return True
		if tp in (tuple,) and self.values[key] == None:
			return True
		# Return value
		return type(self.values[key]) == tp
	
	def save(self):
		""" Saves configuration file """
		with open(self.get_config_file(), "w") as conf:
			conf.write(
				json.dumps(
					self.values,
					sort_keys=True,
					indent=4,
					separators=(',', ': '),
					default=serializer
				)
			)

	def __iter__(self):
		for k in self.values:
			yield k
	
	def get(self, key):
		return self.values[key]
	
	def set(self, key, value):
		self.values[key] = value
		self.save()
	
	__getitem__ = get
	__setitem__ = set
	
	def __delitem__(self, key):
		del self.values[key]
	
	def __contains__(self, key):
		""" Returns true if there is such value """
		return key in self.values

def serializer(obj):
	""" Handles serialization where json can't do it by itself """
	if hasattr(obj, "isoformat"):
		# datetime object
		return obj.isoformat()
	raise TypeError("Can't serialize object of type %s" % (type(obj),))

def Configuration(*a, **b):
	if IS_WINDOWS and not is_portable():
		from syncthing_gtk.windows import WinConfiguration
		return WinConfiguration(*a, **b)
	return _Configuration(*a, **b)
Configuration.REQUIRED_KEYS = _Configuration.REQUIRED_KEYS
Configuration.WINDOWS_OVERRIDE = _Configuration.WINDOWS_OVERRIDE


def migrate_fs_watch(stgtk_config, st_config):
	"""
	Migrates filesystem watch config from ST-GTK configuration
	to Syncthing configuration and posts it to daemon.
	
	Returns True if anything was changed.
	
	Called automatically if old fs watch setting is found in ST-GTK config.
	"""
	# TODO: This can be removed later
	if "use_inotify" not in stgtk_config:
		return False
	changed = False
	folder_by_id = { x["id"]: x for x in st_config['folders'] }
	for rid in stgtk_config["use_inotify"]:
		if rid in folder_by_id:
			folder_by_id[rid]["fsWatcherDelayS"] = 10
			folder_by_id[rid]["fsWatcherEnabled"] = True
			changed = True
	del stgtk_config["use_inotify"]
	return changed



#!/usr/bin/env python3
"""
Syncthing-GTK - Ignore Pattern Editor
"""


from syncthing_gtk.tools import _ # gettext function
from syncthing_gtk.uibuilder import UIBuilder
import os, logging
log = logging.getLogger("IgnoreEditor")

class IgnoreEditor(object):
	""" Standard looking about dialog """
	def __init__(self, app, rid, file_location):
		# Store stuff
		self.app = app
		self.rid  = rid
		self.file_location = file_location
		# Load UI
		self.setup_widgets()
		
	
	def __getitem__(self, name):
		""" Convince method that allows widgets to be accessed via self["widget"] """
		return self.builder.get_object(name)
	
	def show(self, parent=None):
		if not parent is None:
			self["dialog"].set_transient_for(parent)
		self["dialog"].show_all()
	
	def close(self, *a):
		self["dialog"].set_visible(False)
		self["dialog"].destroy()
	
	def setup_widgets(self):
		# Load glade file
		self.builder = UIBuilder()
		self.builder.add_from_file(os.path.join(self.app.gladepath, "ignore-editor.glade"))
		self.builder.connect_signals(self)
		self["lblLocation"].set_markup(
			'%s <a href="file://%s">%s</a>' % (
			_("File location:"),
			os.path.join(os.path.expanduser(self.file_location), ".stignore"),
			os.path.join(self.file_location, ".stignore")
			)
		)
	
	def on_dialog_response(self, *a):
		self.close()
	
	def cb_btClose_clicked(self, *a):
		self.close()
	
	def on_lblLocation_activate_link(self, *a):
		# Called when user clicks on file location link. Clicking there
		# should open .stignore file in default text editor, allowing
		# user to edit it there. Saving file from this dialog afterwards
		# would overwrite his changes, so dialog closes itself to
		# prevent that from happening
		self.close()
		
	def btSave_clicked_cb(self, *a):
		start_iter = self["tbPatterns"].get_start_iter()
		end_iter = self["tbPatterns"].get_end_iter()
		text = self["tbPatterns"].get_text(start_iter, end_iter, True)
		self["tvPatterns"].set_sensitive(False)
		self["btSave"].set_sensitive(False)
		# TODO: Expect error and create appropriate callback for it
		self.app.daemon.write_stignore(self.rid, text, self.close, self.close)
	
	def load(self):
		self.app.daemon.read_stignore(self.rid, self.cb_data_loaded, self.cb_data_failed)
	
	def cb_data_failed(self, *a):
		# This should be next to impossible, so simply closing dialog
		# should be enough of "solution"
		log.error("Failed to load .stignore data: %s", a)
		self.close()
	
	def cb_data_loaded(self, text):
		self["tbPatterns"].set_text(text)
		self["tvPatterns"].grab_focus()
		self["tvPatterns"].set_sensitive(True)
		self["btSave"].set_sensitive(True)



#!/usr/bin/env python3

from distutils.core import setup
from distutils.command.build_py import build_py
from subprocess import Popen, PIPE
import glob, os
APP_ICON_SIZES = (16, 24, 32, 64, 128, 256)
SI_ICON_SIZES = (16, 24, 32)

def get_version():
	"""
	Returns current package version using git-describe or examining
	path. If both methods fails, returns 'unknown'.
	"""
	try:
		p = Popen(['git', 'describe', '--tags', '--match', 'v*'], stdout=PIPE)
		version = p.communicate()[0].strip("\n\r \t")
		if p.returncode != 0:
			raise Exception("git-describe failed")
		return version
	except: pass
	# Git-describe method failed, try to guess from working directory name
	path = os.getcwd().split(os.path.sep)
	version = 'unknown'
	while len(path):
		# Find path component that matches 'syncthing-gui-vX.Y.Z'
		if path[-1].startswith("syncthing-gui-") or path[-1].startswith("syncthing-gtk-"):
			version = path[-1].split("-")[-1]
			if not version.startswith("v"):
				version = "v%s" % (version,)
			break
		path = path[0:-1]
	return version

class BuildPyEx(build_py):
	""" Little extension to install command; Allows --nostdownloader argument """
	user_options = build_py.user_options + [
		# Note to self: use
		# # ./setup.py build_py --nostdownloader install
		# to enable this option
		#
		('nostdownloader', None, 'prevents installing StDownloader module; disables autoupdate capability'),
		('nofinddaemon', None, 'prevents installing FindDaemonDialog module; always uses only default path to syncthig binary'),
	]
	
	def run(self):
		build_py.run(self)
	
	def initialize_options(self):
		build_py.initialize_options(self)
		self.nostdownloader = False
		self.nofinddaemon = False
	
	@staticmethod
	def _remove_module(modules, to_remove):
		for i in modules:
			if i[1] == to_remove:
				modules.remove(i)
				return
		
	
	def find_package_modules(self, package, package_dir):
		rv = build_py.find_package_modules(self, package, package_dir)
		if self.nostdownloader:
			BuildPyEx._remove_module(rv, "stdownloader")
		if self.nofinddaemon:
			BuildPyEx._remove_module(rv, "finddaemondialog")
		return rv

def find_mos(parent, lst=[]):
	for f in os.listdir(parent):
		fp = os.path.join(parent, f)
		if os.path.isdir(fp):
			find_mos(fp, lst)
		elif fp.endswith(".mo"):
			lst += [ fp ]
	return lst

if __name__ == "__main__" : 
	data_files = [
		('share/syncthing-gtk', glob.glob("glade/*.glade") ),
		('share/syncthing-gtk', glob.glob("scripts/syncthing-plugin-*.py") ),
		('share/syncthing-gtk/icons', [
				"icons/%s.svg" % x for x in (
					'add_node', 'add_repo', 'address',
					'announce', 'clock', 'compress', 'cpu', 'dl_rate',
					'eye', 'folder', 'global', 'home', 'ignore', 'lock',
					'ram', 'shared', 'show_id', 'show_id', 'sync', 'thumb_up',
					'up_rate', 'version', 'rescan'
			)] + [
				"icons/%s.png" % x for x in (
					'restart', 'settings', 'shutdown', "st-gtk-logo"
			)]),
		('share/man/man1', glob.glob("doc/*") ),
		('share/icons/hicolor/64x64/emblems', glob.glob("icons/emblem-*.png") ),
		('share/pixmaps', ["icons/syncthing-gtk.png"]),
		('share/applications', ['syncthing-gtk.desktop'] ),
		('share/metainfo', ['me.kozec.syncthingtk.appdata.xml'] ),
	] + [
		(
			'share/icons/hicolor/%sx%s/apps' % (size,size),
			glob.glob("icons/%sx%s/apps/*" % (size,size))
		) for size in APP_ICON_SIZES 
	] + [
		(
			'share/icons/hicolor/%sx%s/status' % (size,size),
			glob.glob("icons/%sx%s/status/*" % (size,size))
		) for size in SI_ICON_SIZES
	] + [
		("share/" + os.path.split(x)[0], (x,)) for x in find_mos("locale/")
	]
	setup(
		name = 'syncthing-gtk',
		version = get_version(),
		description = 'GTK3 GUI for Syncthing',
		url = 'https://github.com/syncthing/syncthing-gtk',
		packages = ['syncthing_gtk'],
		install_requires = (
		    'python-dateutil',
		    'bcrypt',
		),
		data_files = data_files,
		scripts = [ "scripts/syncthing-gtk" ],
		cmdclass = { 'build_py': BuildPyEx },
	)



#!/c/Python27/python.exe
"""
Update Chocolatey package from GitHub releases information.

Requirements:
- Being on Windows
- Having Chocolatey installed (https://chocolatey.org/)
- Having warmup and nuget.commandline installed ("cinst warmup nuget.commandline")
- Having the API key configured ("nuget SetApiKey [API_KEY_HERE] -source http://chocolatey.org/" "https://chocolatey.org/account")
- Being a maintainer of the syncthing-gtk package on Chocolatey
"""


from __future__ import unicode_literals, print_function

import re, os, json
try:
	from urllib import request # Py3
except ImportError:
	import urllib2 as request  # Py2
	from io import open


print("Retrieving last version...")

releasesString = request.urlopen("https://api.github.com/repos/syncthing/syncthing-gtk/releases").read().decode('utf-8')
releases = json.loads(releasesString)

lastRelease = releases[0] # Improve if needed
version = ''
url = ''
releaseNotes = ''

version = lastRelease['name'].replace('v', '', )
releaseNotes = lastRelease['body'].replace('\r', '').replace(':\n-', ':\n\n-')

for asset in lastRelease['assets']:
	if re.match(r'.+win32-full-installer.exe', asset['name']):
		# url = "https://cdn.rawgit.com/syncthing/syncthing-gtk/releases/download/"+lastRelease['name']+"/"+asset['name']
		url = asset['browser_download_url']
assert(url != ''), "ERR No fitting script found"


print("Found version", version)

print("Updating files...")

nuspecFile = open("syncthing-gtk.nuspec", "r", encoding="utf8")
nuspecString = nuspecFile.read()
nuspecFile.close()

nuspecString = re.sub(r'<version>.*</version>', '<version>'+version+'</version>', nuspecString)
nuspecString = re.sub(r'<releaseNotes>[\w\W]*</releaseNotes>', '<releaseNotes>'+releaseNotes+'</releaseNotes>', nuspecString)

nuspecFile = open("syncthing-gtk.nuspec", "w", encoding="utf8")
print(nuspecString, file=nuspecFile, end="")
nuspecFile.close()

chocolateyInstallFile = open("tools/chocolateyInstall.ps1", "r", encoding="utf8")
chocolateyInstallString = chocolateyInstallFile.read()
chocolateyInstallFile.close()

chocolateyInstallString = re.sub(r'\$url ?= ?\'.*\'\n', '$url = \''+url+'\'\n', chocolateyInstallString)

chocolateyInstallFile = open("tools/chocolateyInstall.ps1", "w", encoding="utf8")
print(chocolateyInstallString, file=chocolateyInstallFile, end="")
chocolateyInstallFile.close()

print("Packaging...")

os.system("cpack")

input("Done! Press [Enter] to push or ^C to cancel pushing")

print("Pushing...")

os.system("cpush syncthing-gtk."+version+".nupkg")

print("Done!")



#!/usr/bin/env python3
"""
Syncthing-GTK - DeviceEditorDialog

Universal dialog handler for all Syncthing settings and editing
"""


from gi.repository import Gtk
from syncthing_gtk.tools import check_device_id
from syncthing_gtk.tools import _ # gettext function
from syncthing_gtk.editordialog import EditorDialog, strip_v
import logging
log = logging.getLogger("DeviceEditor")

COLOR_NEW				= "#A0A0A0"
VALUES = [ "vdeviceID", "vname", "vaddresses", "vcompression",
	"vfolders", "vintroducer"
	]

class DeviceEditorDialog(EditorDialog):
	MESSAGES = {
		# Displayed when device id is invalid
		"vdeviceID" : _("The entered device ID does not look valid. It "
			"should be a 52 character string consisting of letters and "
			"numbers, with spaces and dashes being optional."),
	}
	
	def __init__(self, app, is_new, id=None):
		EditorDialog.__init__(self, app,
			"device-edit.glade",
			"New Device" if is_new else "Edit Device"
			)
		self.id = id
		self.is_new = is_new
	
	#@Overrides
	def get_value(self, key):
		if key == "addresses":
			return ",".join([ strip_v(x) for x in self.values[key]])
		elif key == "deviceID":
			return EditorDialog.get_value(self, key).strip(" \r\n\t")
		elif key == "compression":
			val = EditorDialog.get_value(self, key)
			# For syncthing <= 0.10.25
			if val in (True, "true"):
				return "always"
			elif val in (False, "false"):
				return "never"
			else:
				return val
		else:
			return EditorDialog.get_value(self, key)
	
	#@Overrides
	def set_value(self, key, value):
		if key == "addresses":
			self.values[key] = [ strip_v(x) for x in value.split(",") ]
		else:
			return EditorDialog.set_value(self, key, value)
	
	#@Overrides
	def on_data_loaded(self):
		try:
			if self.is_new:
				self.values = { strip_v(x) : "" for x in VALUES }
				self.set_value("addresses", "dynamic")
				self.set_value("compression", "metadata")
				self.checks = {
					"vdeviceID" : check_device_id,
					}
				if self.id != None:
					# Pre-fill device id, if provided
					self.set_value("deviceID", self.id)
			else:
				self.values = [ x for x in self.config["devices"] if x["deviceID"] == self.id ][0]
		except KeyError as e:
			# ID not found in configuration. This is practically impossible,
			# so it's handled only by self-closing dialog.
			log.exception(e)
			self.close()
			return
		return self.display_values(VALUES)
	
	#@Overrides
	def display_value(self, key, w):
		if key == "vfolders":
			# Even more special case
			rids = [ ]
			# Get list of folders that share this device
			for r in self.config["folders"]:
				for n in r["devices"]:
					if n["deviceID"] == self.id and r["id"] not in rids:
						rids.append(r["id"])
			# Create CheckButtons
			for folder in reversed(sorted(self.app.folders.values(), key=lambda x : x["id"])):
				b = Gtk.CheckButton(folder["path"], False)
				b.set_tooltip_text(folder["id"])
				self["vfolders"].pack_start(b, False, False, 0)
				b.set_active(folder["id"] in rids)
			self["vfolders"].show_all()
		else:
			EditorDialog.display_value(self, key, w)
	
	#@Overrides
	def update_special_widgets(self, *a):
		self["vdeviceID"].set_sensitive(self.is_new)
	
	#@Overrides
	def on_save_requested(self):
		self.store_values(VALUES)
		if self.is_new:
			# Add new dict to configuration (edited dict is already there)
			self.config["devices"].append(self.values)
		# Post configuration back to daemon
		self.post_config()
	
	#@Overrides
	def store_value(self, key, w):
		if key == "vaddresses":
			addresses = [
				x.strip() if "://" in x or x.strip() == "dynamic" else "tcp://%s" % (x.strip(),)
				for x in w.get_text().split(",") ]
			self.set_value("addresses", ",".join(addresses))
		elif key == "vfolders":
			# Generate dict of { folder_id : bool } where bool is True if
			# folder should be shared with this device
			folders = {}
			for b in self["vfolders"].get_children():
				folders[b.get_tooltip_text()] = b.get_active()
			# Go over all Folders/<folder>/Devices/<device> keys in config
			# and set them as needed
			nid = self.get_value("deviceID")
			for r in self.config["folders"]:
				rid = r["id"]
				found = False
				for n in r["devices"]:
					if n["deviceID"] == nid:
						if not rid in folders or not folders[rid]:
							# Remove this /<device> key (unshare folder with device)
							r["devices"].remove(n)
							break
						found = True
				if (not found) and (rid in folders) and folders[rid]:
					# Add new /<device> key (share folder with device)
					r["devices"].append({
					   "addresses" : None,
					   "deviceID" : nid,
					   "name" : "",
					   "certName" : "",
					   "compression" : "metadata"
						})
		else:
			EditorDialog.store_value(self, key, w)
	
	#@Overrides
	def on_saved(self):
		self.close()
